import hjson

from .htmlcomponents import *
from addict import Dict
import itertools
from urllib.parse import quote


# TODO: May need to call chart.reflow() on resize
# TODO: Handle formatter functions, for example in dataLabels and others.
# TODO: Add support for more events like drilldown

# If width of chart not specified it defaults to 600px
# A JavaScript date is fundamentally specified as the number of milliseconds that have elapsed since midnight on January 1, 1970, UTC
# https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/UTC


def make_pairs_list(x_data, y_data):
    return list(map(list, itertools.zip_longest(x_data, y_data)))


class HighCharts(JustpyBaseComponent):
    # Highcharts.getOptions().colors
    highcharts_colors = [
        "#7cb5ec",
        "#434348",
        "#90ed7d",
        "#f7a35c",
        "#8085e9",
        "#f15c80",
        "#e4d354",
        "#2b908f",
        "#f45b5b",
        "#91e8e1",
    ]

    # Theme is One of ['avocado', 'dark-blue', 'dark-green', 'dark-unica', 'gray',
    # 'grid-light', 'grid', 'high-contrast-dark', 'high-contrast-light', 'sand-signika', 'skies', 'sunset']
    # but is set at the WebPage level. All charts on same page have the same theme.
    # Example: wp.highcharts_theme = 'grid'

    vue_type = "chart"  # The corresponding Vue component

    chart_types = [
        "columnrange",
        "cylinder",
        "dependencywheel",
        "errorbar",
        "funnel",
        "funnel3d",
        "gauge",
        "heatmap",
        "histogram",
        "item",
        "line",
        "networkgraph",
        "organization",
        "packedbubble",
        "pareto",
        "pie",
        "polygon",
        "pyramid",
        "pyramid3d",
        "sankey",
        "scatter",
        "scatter3d",
        "solidgauge",
        "spline",
        "streamgraph",
        "sunburst",
        "tilemap",
        "timeline",
        "treemap",
        "variablepie",
        "variwide",
        "vector",
        "venn",
        "waterfall",
        "windbarb",
        "wordcloud",
        "xrange",
    ]

    def __init__(self, **kwargs):
        self.options = Dict()
        self.stock = False
        self.use_cache = True
        self.classes = ""
        self.style = ""
        self.show = True
        self.event_propagation = True
        self.pages = {}
        self.tooltip_fixed = False
        self.tooltip_x = -40
        self.tooltip_y = 40
        self.tooltip_debounce = 100  # Default is 100 ms
        self.update_animation = True  # Whether to animate changes when chart is updated
        self.update_create = False  # Whether to create new chart on update, if false current chart is updated
        kwargs["temp"] = False  # Force an id to be assigned to chart
        super().__init__(**kwargs)
        for k, v in kwargs.items():
            self.__setattr__(k, v)
        self.allowed_events = [
            "tooltip",
            "point_click",
            "point_select",
            "point_unselect",
            "series_hide",
            "series_show",
            "series_click",
            "zoom_x",
            "zoom_y",
        ]
        for e in self.allowed_events:
            for prefix in ["", "on", "on_"]:
                if prefix + e in kwargs.keys():
                    fn = kwargs[prefix + e]
                    if isinstance(fn, str):
                        fn_string = f"def oneliner{self.id}(self, msg):\n {fn}"
                        exec(fn_string)
                        self.on(e, locals()[f"oneliner{self.id}"])
                    else:
                        self.on(e, fn)
                    break
        if type(self.options) != Dict:
            self.options = Dict(self.options)
        if "series" not in self.options:
            self.options.series = []
        for com in ["a", "add_to"]:
            if com in kwargs.keys():
                kwargs[com].add_component(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(id: {self.id}, vue_type: {self.vue_type}, chart options: {self.options})"

    def __setattr__(self, key, value):
        if key == "options":
            if isinstance(value, str):
                self.load_json(value)
            else:
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    async def chart_update(self, update_dict, websocket):
        # https://api.highcharts.com/class-reference/Highcharts.Chart#update
        await websocket.send_json(
            {"type": "chart_update", "data": update_dict, "id": self.id}
        )
        # So the page itself does not update, only the tooltip, return True not None
        return True

    async def tooltip_update(self, tooltip, websocket):
        await websocket.send_json(
            {"type": "tooltip_update", "data": tooltip, "id": self.id}
        )
        # So the page itself does not update, only the tooltip, return True not None
        return True

    async def draw_crosshair(self, point_list, websocket):
        """
        point_list is list of dictionaries  whose keys are:
        'id': the chart id
        'series': the series index
        'point': the point index
        Values are  all integers
        Example:
         {'id': chart_id, 'series': msg.series_index, 'point': msg.point_index}
        """
        await websocket.send_json({"type": "draw_crosshair", "data": point_list})
        # Return True not None so that the page does not update
        return True

    async def select_point(self, point_list, websocket):
        """
        point_list is list of of dictionaries  whose keys are:
        'id': the chart id
        'series': the series index
        'point': the point index
        Values are  all integers
        Example:
         {'id': chart_id, 'series': msg.series_index, 'point': msg.point_index}
        """
        await websocket.send_json({"type": "select_point", "data": point_list})
        # Return True not None so that the page does not update
        return True

    def add_to_page(self, wp: WebPage):
        wp.add_component(self)

    def add_to(self, *args):
        for c in args:
            c.add_component(self)

    def react(self, data):
        pass

    def load_json(self, options_string):
        self.options = Dict(hjson.loads(options_string.encode("ascii", "ignore")))
        return self.options

    def load_json_from_file(self, file_name):
        with open(file_name, "r") as f:
            self.options = Dict(hjson.loads(f.read().encode("ascii", "ignore")))
        return self.options

    def convert_object_to_dict(self):

        d = {}
        d["vue_type"] = self.vue_type
        d["id"] = self.id
        d["stock"] = self.stock
        d["use_cache"] = self.use_cache
        d["show"] = self.show
        d["classes"] = self.classes
        d["style"] = self.style
        d["event_propagation"] = self.event_propagation
        d["def"] = self.options
        d["events"] = self.events
        d["tooltip_fixed"] = self.tooltip_fixed
        d["tooltip_x"] = self.tooltip_x
        d["tooltip_y"] = self.tooltip_y
        d["tooltip_debounce"] = self.tooltip_debounce
        d["update_animation"] = self.update_animation
        d["update_create"] = self.update_create
        return d


class HighStock(HighCharts):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stock = True


class Histogram(HighCharts):
    _options = """
{
    title: {
        text: 'Highcharts Histogram'
    },
    xAxis: [{
        title: { text: 'Data' },
        alignTicks: false
    }, {
        title: { text: 'Histogram' },
        alignTicks: false,
        opposite: true
    }],

    yAxis: [{
        title: { text: 'Data' }
    }, {
        title: { text: 'Histogram' },
        opposite: true
    }],

    series: [{
        name: 'Histogram',
        type: 'histogram',
        xAxis: 1,
        yAxis: 1,
        baseSeries: 's1',
        zIndex: -1
    }, {
        name: 'Data',
        type: 'scatter',
        data: [],
        id: 's1',
        marker: {
            radius: 1.5
        }
    }]
}

    """

    def __init__(self, data, **kwargs):
        super().__init__(**kwargs)
        self.load_json(self._options)
        self.options.series[1].data = list(data)


class Pie(HighCharts):
    _options = """
            {
                chart: {
                    title: {
                        text: 'Pie Chart'
                        }
                    },
                    tooltip: {
                        pointFormat: '{series.name}: <b>{point.percentage:.1f}%</b>'
                    },
                    plotOptions: {
                        pie: {
                            allowPointSelect: true,
                            cursor: 'pointer',
                            dataLabels: {
                                enabled: true,
                                format: '<b>{point.name}</b>: {point.percentage:.1f} %',
                            }
                        }
                    },
                    series: []
            }
                """

    def __init__(self, data, **kwargs):
        self.labels = []
        super().__init__(**kwargs)
        self.load_json(self._options)
        series = Dict()
        series.type = "pie"
        series_data = []
        series.data = series_data
        for i, value in enumerate(data):
            c = Dict()
            try:
                c.name = self.labels[i]
            except:
                c.name = str(value)
            c.y = value
            series_data.append(c)
        self.options.series.append(series)


class PieSemiCircle(HighCharts):
    _options = """
            {
                chart: {
                    plotBackgroundColor: null,
                    plotBorderWidth: null,
                    plotShadow: false,
                },
                title: {
            text: 'SemiCircle Chart',
            align: 'center',
            verticalAlign: 'middle',
            y: 40
        },
                tooltip: {
                    pointFormat: '{series.name}: <b>{point.percentage:.1f}%</b>'
                },
                plotOptions: {
            pie: {
                dataLabels: {
                    enabled: true,
                    distance: -50,
                    style: {
                        fontWeight: 'bold',
                        color: 'white'
                    }
                },
                startAngle: -90,
                endAngle: 90,
                center: ['50%', '75%'],
                size: '110%'
            }
        },
                series: []
            }
            """

    def __init__(self, data, **kwargs):
        self.labels = []
        super().__init__(**kwargs)
        self.load_json(self._options)
        series = Dict()
        series.type = "pie"
        series.innerSize = "60%"
        series_data = []
        series.data = series_data
        for i, value in enumerate(data):
            c = Dict()
            try:
                c.name = self.labels[i]
            except:
                c.name = str(value)
            c.y = value
            series_data.append(c)
        self.options.series.append(series)


class Scatter(HighCharts):
    _options = """
    {
    chart: {
        type: 'scatter',
        zoomType: 'xy'
    },
    title: {
        text: 'Scatter Chart'
    },
     xAxis: {
        title: {
            enabled: true,
            text: 'x'
        },
        startOnTick: false,
        endOnTick: true,
        showLastLabel: true
    },
    yAxis: {
        title: {
            text: 'y'
        }
    },
    series:[]
    }
    """

    def __init__(self, x, y, **kwargs):
        super().__init__(**kwargs)
        self.load_json(self._options)
        s = Dict()
        s.data = list(zip(x, y))
        self.options.series.append(s)


# --------------------------------------------------------------------
# matplotlib related objects

try:
    import matplotlib.pyplot as plt

    _has_matplotlib = True
    import io
except:
    _has_matplotlib = False

if _has_matplotlib:

    class Matplotlib(Div):
        def __init__(self, **kwargs):
            self.figure = plt.gcf()
            super().__init__(**kwargs)
            self.set_figure(self.figure)

        def set_figure(self, fig=None):
            if not fig:
                fig = self.figure
            plt.figure(fig.number)
            output = io.StringIO()
            plt.savefig(output, format="svg")
            # The 'white-space:pre' causes problems in placement of chart on page
            self.inner_html = output.getvalue().replace(
                "*{stroke-linecap:butt;stroke-linejoin:round;white-space:pre;}",
                "*{stroke-linecap:butt;stroke-linejoin:round;}",
            )
            output.close()
            return self.inner_html


# --------------------------------------------------------------------
# deck.gl related objects

s = """
<div style="width:100%;"><div style="position:relative;width:100%;height:0;padding-bottom:60%;"><iframe src="about:blank" style="position:absolute;width:100%;height:100%;left:0;top:0;border:none !important;" data-html={} onload="this.contentDocument.open();this.contentDocument.write(    decodeURIComponent(this.getAttribute('data-html')));this.contentDocument.close();" allowfullscreen webkitallowfullscreen mozallowfullscreen></iframe></div></div>
"""

try:
    import pydeck as pdk

    _has_pydeck = True
except:
    _has_pydeck = False

if _has_pydeck:

    class PyDeckFrame(Iframe):
        vue_type = "iframejp"

        def __init__(self, **kwargs):
            self.deck = None
            self.srcdoc = None
            kwargs["temp"] = False  # Force an id to be assigned to chart
            self.view_delay = (
                0  # Additional delay in ms for display after frame has loaded
            )
            self.transition_duration = (
                0.1  # Duration of transition between frames in seconds
            )
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            self.srcdoc = quote(self.deck.to_html(as_string=True))
            d = super().convert_object_to_dict()
            d["view_delay"] = self.view_delay
            d["transition_duration"] = self.transition_duration
            return d

    class PyDeck(Div):
        vue_type = "deckgl"

        def __init__(self, **kwargs):
            self.use_cache = False
            self.deck = None
            kwargs["temp"] = False  # Force an id to be assigned to chart
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            d = {}
            d["vue_type"] = self.vue_type
            d["id"] = self.id
            d["use_cache"] = self.use_cache
            d["show"] = self.show
            d["classes"] = self.classes
            d["style"] = self.style
            d["event_propagation"] = self.event_propagation
            d["deck"] = self.deck.to_json()
            d["events"] = self.events
            d["mapbox_key"] = self.deck.mapbox_key
            return d


try:
    import altair as alt

    _has_altair = True
except:
    _has_altair = False

if _has_altair:

    class AltairChart(Div):

        vue_type = "altairjp"

        def __init__(self, **kwargs):
            self.use_cache = False
            self.chart = None
            self.options = {}
            self.vega_source = None
            # self.style = 'width:100%; height: 500px'
            kwargs["temp"] = False  # Force an id to be assigned to chart
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            d = {}
            d["vue_type"] = self.vue_type
            d["id"] = self.id
            d["use_cache"] = self.use_cache
            d["show"] = self.show
            d["classes"] = self.classes
            d["style"] = self.style
            d["event_propagation"] = self.event_propagation
            if self.vega_source:
                d["vega_source"] = json.dumps(self.vega_source)
            else:
                d["vega_source"] = self.chart.to_json()
            d["events"] = self.events
            d["options"] = self.options
            return d


try:
    import plotly

    _has_plotly = True
except:
    _has_plotly = False

if _has_plotly:

    class PlotlyChart(Div):

        vue_type = "plotlyjp"

        def __init__(self, **kwargs):
            self.use_cache = False
            self.chart = None
            self.chart_dict = {}
            self.config = {}
            kwargs["temp"] = False  # Force an id to be assigned to chart
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            d = {}
            d["vue_type"] = self.vue_type
            d["id"] = self.id
            d["use_cache"] = self.use_cache
            d["show"] = self.show
            d["classes"] = self.classes
            d["style"] = self.style
            d["event_propagation"] = self.event_propagation
            if self.chart:
                d["chart"] = self.chart.to_json()
            else:
                d["chart"] = self.chart_dict
            d["events"] = self.events
            d["config"] = self.config
            return d


try:
    import bokeh

    _has_bokeh = True
except:
    _has_bokeh = False

if _has_bokeh:

    class BokehChart(Div):

        vue_type = "bokehjp"

        def __init__(self, **kwargs):
            self.use_cache = False
            self.chart = None
            self.chart_dict = {}
            self.config = {}
            kwargs["temp"] = False  # Force an id to be assigned to chart
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            d = {}
            d["vue_type"] = self.vue_type
            d["id"] = self.id
            d["use_cache"] = self.use_cache
            d["show"] = self.show
            d["classes"] = self.classes
            d["style"] = self.style
            d["event_propagation"] = self.event_propagation
            if self.chart:
                d["chart"] = json.dumps(bokeh.embed.standalone.json_item(self.chart))
            else:
                d["chart"] = self.chart_dict
            d["events"] = self.events
            d["config"] = self.config
            return d


try:
    import folium

    _has_folium = True
except:
    _has_folium = False

if _has_folium:

    class FoliumChart(Div):
        def __init__(self, **kwargs):
            self.use_cache = False
            self.chart = None
            kwargs["temp"] = False  # Force an id to be assigned to chart
            super().__init__(**kwargs)
            self.inner_div = Div(a=self)

        def convert_object_to_dict(self):
            if self.chart:
                self.inner_html = self.chart._repr_html_()
            d = super().convert_object_to_dict()
            return d



"""
Created on 2022-08-29

@author: wf
"""
import urllib.request
import os
import gzip
import shutil
from pathlib import Path


class Download:
    """
    Utility functions for downloading data
    """

    @staticmethod
    def get_cache_path():
        home = str(Path.home())
        cachedir = f"{home}/.justpy"
        return cachedir

    @staticmethod
    def get_url_content(url: str):
        with urllib.request.urlopen(url) as url_response:
            content = url_response.read().decode()
            return content

    @staticmethod
    def get_file_content(path: str):
        with open(path, "r") as file:
            content = file.read()
            return content

    @staticmethod
    def needs_download(file_path: str, force: bool = False) -> bool:
        """
        Check if a download of the given file_path is necessary that is the file
        does not exist has a size of zero or the download should be forced

        Args:
            file_path(str): the path of the file to be checked
            force(bool): True if the result should be forced to True

        Return:
            bool: True if  a download for this file needed
        """
        if not os.path.isfile(file_path):
            result = True
        else:
            stats = os.stat(file_path)
            size = stats.st_size
            result = force or size == 0
        return result

    @staticmethod
    def download_file(
        url: str, file_name: str, target_directory: str, force: bool = False
    ):
        if Download.needs_download(target_directory, force):
            file_path = f"{target_directory}/{file_name}"
            urllib.request.urlretrieve(url, file_path)

    @staticmethod
    def download_backup_file(
        url: str, file_name: str, target_directory: str, force: bool = False
    ):
        """
        Downloads from the given url the zip-file and extracts the file corresponding to the given file_name.

        Args:
            url: url linking to a downloadable gzip file
            file_name: Name of the file that should be extracted from gzip file
            target_directory(str): download the file this directory
            force (bool): True if the download should be forced

        Returns:
            Name of the extracted file with path to the backup directory
        """
        extract_to = f"{target_directory}/{file_name}"
        # we might want to check whether a new version is available
        if Download.needs_download(extract_to, force=force):
            if not os.path.isdir(target_directory):
                os.makedirs(target_directory)
            zipped = f"{extract_to}.gz"
            print(f"Downloading {zipped} from {url} ... this might take a few seconds")
            urllib.request.urlretrieve(url, zipped)
            print(f"Unzipping {extract_to} from {zipped}")
            with gzip.open(zipped, "rb") as gzipped:
                with open(extract_to, "wb") as unzipped:
                    shutil.copyfileobj(gzipped, unzipped)
                print("Extracting completed")
            if not os.path.isfile(extract_to):
                raise (f"could not extract {file_name} from {zipped}")
        return extract_to



'''
Created on 2022-09-14

@author: wf
'''
import asyncio
from selenium.webdriver.common.by import By
from selenium.webdriver.support.select import Select
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from tests.base_selenium_test import BaseSeleniumTest

class TestDatabaseExample(BaseSeleniumTest):
    """
    testing database example
    """

    async def setUp(self):
        await super().setUp(port=8127)
        
    def findElement(self,condition,timeout=4.0):
        element_present = EC.presence_of_element_located(condition)
        element=WebDriverWait(self.browser, timeout).until(element_present)
        return element
        
    async def testDatabaseExample(self):
        """
        test database example from grids_tutorial/database.md
        """
        self.browser = await self.getBrowserForDemo()
        from examples.tutorial.db_test import db_test
        await self.server.start(db_test)
        url = self.server.get_url("/")
        self.browser.get(url)
        await asyncio.sleep(self.server.sleep_time)
        
        selectInput=self.findElement((By.CLASS_NAME,"q-select__autocomplete-input"))
        #select=Select(selectElement)
        #select.selectByIndex(2)
        #selectInput.clear()
        #selectInput.send_keys("artists")
        await asyncio.sleep(3)
        self.browser.close()
        await asyncio.sleep(self.server.sleep_time)
        await self.server.stop()


"""
Created on 2022-08-24

@author: wf
"""
import justpy as jp
from tests.base_server_test import BaseAsynctest

class TestJustpyCore(BaseAsynctest):
    """
    Tests for Justpy Core features
    """

    async def setUp(self):
        await BaseAsynctest.setUp(self, port=8122, debug=True)

    async def wp_to_test(self):
        """
        the example Webpage under test
        """
        wp = jp.WebPage()
        _d = jp.Div(a=wp)
        return wp

    async def testWp(self):
        """'
        test the webpage asynchronously
        """
        await self.server.start(self.wp_to_test)
        status, rawhtml = await self.getResponseHtml()
        self.assertEqual(200, status)
        html = rawhtml.decode("utf8")
        debug = True
        if debug:
            print(html)
            with open("/tmp/justpy.html", "w") as html_file:
                print(html, file=html_file)



# Justpy Tutorial demo inner_demo from docs/tutorial/working_with_html.md
import justpy as jp

my_html = """
    <div>
    <p class="m-2 p-2 text-red-500 text-xl">Paragraph 1</p>
    <p class="m-2 p-2 text-blue-500 text-xl">Paragraph 2</p>
    <p class="m-2 p-2 text-green-500 text-xl">Paragraph 3</p>
    </div>
    """

def inner_demo():
    wp = jp.WebPage()
    d = jp.Div(a=wp, classes='m-4 p-4 text-3xl')
    d.inner_html = '<pre>Hello there. \n How are you?</pre>'
    jp.Div(a=wp, inner_html=my_html)
    for color in ['red', 'green', 'blue', 'pink', 'yellow', 'teal', 'purple']:
        jp.Div(a=wp, inner_html=f'<p class="ml-2 text-{color}-500 text-3xl">{color}</p>')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("inner_demo",inner_demo)



# Justpy Tutorial demo grid_test17 from docs/grids_tutorial/pandas.md
import justpy as jp
import pandas as pd

wm_df = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)

def grid_test17():
    wp = jp.WebPage()
    wm_df.jp.ag_grid(a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("grid_test17",grid_test17)



# Justpy Tutorial demo shuffle_test from docs/reference/htmlcomponent.md
import justpy as jp
import random

def shuffle_test():
    wp = jp.WebPage()
    main_div = jp.Div(classes='flex flex-wrap m-2 p-2 ', a=wp)
    for i in range(1,101,1):
        jp.Div(text=f'Div {i}', a=main_div, classes='m-2 p-2 text-xl text-white bg-blue-500')
    random.shuffle(main_div.components)
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("shuffle_test",shuffle_test)



# Justpy Tutorial demo input_test from docs/tutorial/custom_components.md
import justpy as jp

# https://quasar.dev/vue-components/date#With-QInput

class QInputDateTime(jp.QInput):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        date_slot = jp.QIcon(name='event', classes='cursor-pointer')
        c2 = jp.QPopupProxy(transition_show='scale', transition_hide='scale', a=date_slot)
        self.date = jp.QDate(mask='YYYY-MM-DD HH:mm', name='date', a=c2)

        time_slot = jp.QIcon(name='access_time', classes='cursor-pointer')
        c2 = jp.QPopupProxy(transition_show='scale', transition_hide='scale', a=time_slot)
        self.time = jp.QTime(mask='YYYY-MM-DD HH:mm', format24h=True, name='time', a=c2)

        self.date.parent = self
        self.time.parent = self
        self.date.value = self.value
        self.time.value = self.value
        self.prepend_slot = date_slot
        self.append_slot = time_slot
        self.date.on('input', self.date_time_change)
        self.time.on('input', self.date_time_change)
        self.on('input', self.input_change)

    @staticmethod
    def date_time_change(self, msg):
        self.parent.value = self.value
        self.parent.date.value = self.value
        self.parent.time.value = self.value

    @staticmethod
    def input_change(self, msg):
        self.date.value = self.value
        self.time.value = self.value


def input_test():
    wp = jp.QuasarPage()
    QInputDateTime(filled=True, style='width: 600px', a=wp, classes="q-pa-md", value='')
    QInputDateTime(filled=True, style='width: 600px', a=wp, classes="q-pa-md", value='2020-03-01 12:44')
    QInputDateTime(filled=True, style='width: 600px', a=wp, classes="q-pa-md", value='2021-04-01 14:44')
    QInputDateTime(filled=True, style='width: 600px', a=wp, classes="q-pa-md", value='2022-05-01 18:44')
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test",input_test)



# Justpy Tutorial demo plot_test7 from docs/tutorial/matplotlib.md
import justpy as jp
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# All examples from https://pandas.pydata.org/pandas-docs/stable/user_guide/visualization.html

def plot_test7():
    wp = jp.WebPage()

    jp.Div(text='Series Plot', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    ts = pd.Series(np.random.randn(1000), index=pd.date_range('1/1/2000', periods=1000))
    ts = ts.cumsum()
    ts.plot()
    jp.Matplotlib(a=wp)
    plt.close()

    jp.Div(text='Frame Plot', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    df = pd.DataFrame(np.random.randn(1000, 4), index=ts.index, columns=list('ABCD'))
    df = df.cumsum()
    df.plot()
    jp.Matplotlib(a=wp)
    plt.close()

    jp.Div(text='Histogram', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    df4 = pd.DataFrame({'a': np.random.randn(1000) + 1, 'b': np.random.randn(1000), 'c': np.random.randn(1000) - 1}, columns=['a', 'b', 'c'])
    df4.plot.hist(alpha=0.5)
    jp.Matplotlib(a=wp)
    plt.close()

    jp.Div(text='Multiple Subplots', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    df.diff().hist(color='k', alpha=0.5, bins=50)
    jp.Matplotlib(a=wp)
    plt.close()

    jp.Div(text='Multiple Pie Charts', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    df_pie = pd.DataFrame(3 * np.random.rand(4, 2),index=['a', 'b', 'c', 'd'], columns=['x', 'y'])
    df_pie.plot.pie(subplots=True, figsize=(8, 4))
    jp.Matplotlib(a=wp)
    plt.close()

    # The following example takes some time (about 15 seconds) to generate, so the page loads slowly
    jp.Div(text='Plotting Tools, Scatter Matrix', a=wp, classes='text-white bg-blue-500 text-center text-xl')
    df_scatter = pd.DataFrame(np.random.randn(1000, 4), columns=['a', 'b', 'c', 'd'])
    pd.plotting.scatter_matrix(df_scatter, alpha=0.2, figsize=(10, 10), diagonal='kde')
    jp.Matplotlib(a=wp)
    plt.close()

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("plot_test7",plot_test7)



# Justpy Tutorial demo quasar_rating_test1 from docs/quasar_tutorial/quasar_components.md
import justpy as jp

def quasar_rating_test1():
    wp = jp.QuasarPage(data={'rating': 2})
    d = jp.Div(classes='q-pa-md', a=wp)
    rating_div = jp.Div(classes='q-gutter-y-md column', a=d)
    jp.QRating(size='1.5em', icon='thumb_up', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='2em', icon='favorite_border',color='red-7', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='2.5em', icon='create',color='purple-4', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='3em', icon='pets',color='brown-5', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='4.5em', icon='star_border',color='green-5', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='5em', icon='star_border',icon_selected='star',color='grey', a=rating_div, model=[wp, 'rating'],
               color_selected=['light-green-3', 'light-green-6', 'green', 'green-9', 'green-10'])
    jp.QRating(size='5em', icon='star_border',icon_selected='star',color='green-5', a=rating_div, model=[wp, 'rating'])
    jp.QRating(size='3.5em', max=4,color='red-5', a=rating_div, model=[wp, 'rating'],
               icon=['sentiment_very_dissatisfied', 'sentiment_dissatisfied', 'sentiment_satisfied', 'sentiment_very_satisfied'])
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("quasar_rating_test1",quasar_rating_test1)



# Justpy Tutorial demo input_demo_model4 from docs/tutorial/model_and_data.md
import justpy as jp

def reset_all(self, msg):
    msg.page.data['text'] = ''

async def input_demo_model4(request):
    wp = jp.WebPage(data={'text': 'Initial text'})
    button_classes = 'w-32 m-2 bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded'
    b = jp.Button(text='Reset', click=reset_all, a=wp, classes=button_classes)
    jp.Hr(a=wp)  # Add horizontal like to page
    input_classes = "m-2 bg-gray-200 appearance-none border-2 border-gray-200 rounded xtw-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"
    for i in range(5):
        jp.Input(a=wp, classes=input_classes, placeholder='Please type here', model=[wp, 'text'])
    for i in range(3):
        jp.Div(model=[wp, 'text'], classes='m-2 p-2 h-32 text-xl border-2 overflow-auto', a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_demo_model4",input_demo_model4)



import hjson

from .htmlcomponents import *
from addict import Dict

try:
    import numpy as np
    import pandas as pd
    from pandas.api.types import is_numeric_dtype, is_datetime64_any_dtype
    from pandas import Timestamp

    _has_pandas = True
except:
    _has_pandas = False


class AgGrid(JustpyBaseComponent):
    """
    
    support for 
    https://www.ag-grid.com
    
    see
    https://www.ag-grid.com/javascript-grid-features/
    
    """

    vue_type = "grid"
    default_grid_options = {
        "animateRows": True,
        "rowDragManaged": True,
        "defaultColDef": {
            "filter": True,
            "sortable": True,
            "resizable": True,
            "unSortIcon": True,
            "cellStyle": {"textAlign": "center"},
            "headerClass": "font-bold",
        },
        "columnDefs": [],
        "rowData": [],
    }

    def __init__(self, **kwargs):
        """
        constructor
        """
        self.options = Dict(self.default_grid_options)
        self.classes = ""
        self.style = "height: 99vh; width: 99%; margin: 0.25rem; padding: 0.25rem;"
        self.evaluate = []  # Fields for evaluation
        self.show = True
        self.pages = {}
        self.auto_size = (
            True  # If True, automatically resize columns after load to optimal fit
        )
        self.theme = "ag-theme-balham"  # one of ag-theme-balham, ag-theme-balham-dark, ag-theme-material
        self.html_columns = []
        self.row_data_converter = None  # optional function to convert row data; signature: (rowIndex, colIndex, colKey, cellValue) -> newValue
        kwargs["temp"] = False
        super().__init__(**kwargs)
        for k, v in kwargs.items():
            self.__setattr__(k, v)
        self.allowed_events = []
        if type(self.options) != Dict:
            self.options = Dict(self.options)
        for com in ["a", "add_to"]:
            if com in kwargs.keys():
                kwargs[com].add_component(self)

    def __repr__(self):
        # return f'stam'
        return f"{self.__class__.__name__}(id: {self.id}, vue_type: {self.vue_type}, Grid options: {self.options})"

    def __setattr__(self, key, value):
        if key == "options":
            if isinstance(value, str):
                self.load_json(value)
            else:
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    @property
    def grid_id(self) -> str:
        """
        component id of the grid
        """
        return f"g{self.id}"

    def on(self, event_type, func, **kwargs):
        # https://www.ag-grid.com/javascript-grid-events/
        self.allowed_events.append(event_type)
        super().on(event_type, func, **kwargs)

    def add_to_page(self, wp: WebPage):
        wp.add_component(self)

    def add_to(self, *args):
        for c in args:
            c.add_component(self)

    def react(self, data):
        pass

    def load_json(self, options_string):
        self.options = Dict(hjson.loads(options_string.encode("ascii", "ignore")))
        return self.options

    def load_json_from_file(self, file_name):
        with open(file_name, "r") as f:
            self.options = Dict(hjson.loads(f.read().encode("ascii", "ignore")))
        return self.options

    def load_pandas_frame(self, df):
        """
        load the given pandas dataframe
        
        Args:
            df: the dataframe to load
        """
        assert _has_pandas, f"Pandas not installed, cannot load frame"
        columnDefs = []
        for i in df.columns:
            if is_numeric_dtype(df[i]):
                col_filter = "agNumberColumnFilter"
            elif is_datetime64_any_dtype(df[i]):
                col_filter = "agDateColumnFilter"
            else:
                col_filter = True  # Use default filter
            columnDefs.append(Dict({"field": i, "filter": col_filter}))
        # Change NaN and similar to None for JSON compatibility
        lod = (
            df.replace([np.inf, -np.inf], [sys.float_info.max, -sys.float_info.max])
            .where(pd.notnull(df), None)
            .to_dict("records")
        )
        self.load_lod(lod=lod,columnDefs=columnDefs)
        
    def load_lod(self,lod:list,columnDefs:list=[]):
        """
        load the given list of dicts
        
        Args:
            lod(list): a list of dicts to be loaded into the grid
            columnDefs(list): a list of column definitions
        """
        self.options.columnDefs=columnDefs
        self.options.rowData=lod

    async def run_api(self, command, page):
        await page.run_javascript(f"""cached_grid_def[{self.grid_id}].api.{command}""")

    async def select_all_rows(self, page):
        await page.run_javascript(
            f"""cached_grid_def[{self.grid_id}].api.selectAll()"""
        )

    async def deselect_rows(self, page):
        await page.run_javascript(
            f"""cached_grid_def[{self.grid_id}].api.deselectAll()"""
        )

    async def apply_transaction(self, transaction, page):
        await page.run_javascript(
            f"""cached_grid_def[{self.grid_id}].api.applyTransaction({transaction.__repr__()})"""
        )

    def convert_object_to_dict(self):
        """
        convert object to dict
        """
        d = dict()
        d["vue_type"] = self.vue_type
        d["id"] = self.id
        d["show"] = self.show
        d["classes"] = self.classes + " " + self.theme
        d["style"] = self.style
        options = self.options.deepcopy()
        rows = options.get("rowData", [])
        for row_idx, row_dict in enumerate(rows):
            for col_idx, (col_key, val) in enumerate(row_dict.items()):
                if callable(self.row_data_converter):
                    val = self.row_data_converter(row_idx, col_idx, col_key, val)
                    row_dict[col_key] = val
                if _has_pandas and isinstance(val, Timestamp):
                    row_dict[col_key] = str(val)
        d["def"] = options
        d["auto_size"] = self.auto_size
        d["events"] = self.events
        d["html_columns"] = self.html_columns
        d["evaluate"] = self.evaluate
        return d



import justpy as jp
import pandas as pd

grid_options_1 = """
{
    rowSelection: 'multiple',
    defaultColDef: {
        filter: true,
        sortable: true,
        resizable: true,
        cellStyle: {textAlign: 'center'},
        headerClass: 'font-bold'
    }
}
"""
# df1 = <data>
df1 = {}

def issue_342_grid_test(request):

  # ?? how do you get AG-Grid get api information --->
  async def saveGridState(self, msg):
    wp = msg.page
    wp.column_state = await self.grid.run_api('columnApi.getColumnState()', msg.page)


  wp = jp.WebPage()
  wp.selected_rows = {}  # Dictionary holding selected rows
  wp.column_state = {}   # attempt to save the column state from AG-Grid api
  grid = jp.AgGrid(a=wp, options=grid_options_1, style='height: 400px; width: 800px; margin: 0.25em')

  grid.load_pandas_frame(df1)
  wp.rows_div = jp.Pre(text='Use Save Col State btn to save settings', classes='border text-lg', a=wp)
  btn_state = jp.Button(text='Save Col State', classes=jp.Styles.button_simple+' m-2', a=wp, click=saveGridState)
  btn_state.grid = grid
  return wp

from examples.basedemo import Demo

Demo("Issue 342", issue_342_grid_test)



# Justpy Tutorial demo event_comp_test from docs/tutorial/handling_events.md
import justpy as jp


def event_comp_test():
    wp = jp.WebPage()
    d = jp.Div(text='hello1',
            click='self.text="clicked"',
            mouseenter='self.text="entered"; self.set_class("text-5xl"); msg.page.add(Div(text=f"{len(msg.page)} Additional Div"))',
            mouseleave='self.text="left"; self.set_class("text-xl")',
            classes='text-2xl border p-2 m-2', a=wp )
    d = jp.Div(text='hello2', click='self.text="clicked"', mouseenter='self.text="entered"', classes='text-2xl border p-2 m-2', a=wp )
    d = jp.Div(text='hello3', click='self.text="clicked"', mouseenter='self.text="entered"', classes='text-2xl border p-2 m-2', a=wp )
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("event_comp_test",event_comp_test)



# Justpy Tutorial demo dog_pic3 from docs/tutorial/request_object.md
import justpy as jp

async def get_image2(self, msg):
    r = await jp.get(f'https://dog.ceo/api/breed/{msg.page.breed}/images/random')
    self.src = r['message']

@jp.SetRoute('/breed/{breed}')
async def dog_pic3(request):
    wp = jp.WebPage()
    breed = request.path_params.get('breed', 'papillon')
    wp.breed = breed
    r = await jp.get(f'https://dog.ceo/api/breed/{breed}/images/random')
    img_url = r['message']
    jp.Img(src=img_url, a=wp, classes='m-2 p-2 cursor-pointer', click=get_image2)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("dog_pic3",dog_pic3)



# Justpy Tutorial demo grid_test14 from docs/grids_tutorial/grid_events.md
import justpy as jp
import pandas as pd

df = pd.read_csv('https://raw.githubusercontent.com/fivethirtyeight/data/master/alcohol-consumption/drinks.csv', encoding="ISO-8859-1")

def grid_change(self, msg):
    msg.page.df = jp.read_csv_from_string(msg.data)
    c = msg.page.df.jp.plot(0, [1,2,3,4], kind='column', classes='m-2 p-2 w-2/3 border', title='Alcohol Consumption per Country')
    msg.page.c.options = c.options

def grid_test14():
    wp = jp.WebPage()
    wp.df = df
    wp.c = df.jp.plot(0, [1,2,3,4], kind='column', a=wp, classes='m-2 p-2 border', title='Alcohol Consumption per Country')
    grid = df.jp.ag_grid(a=wp)
    for event_name in ['sortChanged', 'filterChanged', 'columnMoved', 'rowDragEnd']:
        grid.on(event_name, grid_change)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("grid_test14",grid_test14)



"""
Created on 2022-09-02

@author: wf
"""
import asyncio
import fnmatch
import inspect
import json
import logging
import os
import pathlib
import socket
import re
import psutil
import sys
import traceback
import typing
import uuid
from sys import platform
from multiprocessing import Process
from threading import Thread
from starlette.applications import Starlette
from starlette.routing import Route
from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse, JSONResponse,PlainTextResponse, Response
from starlette.templating import Jinja2Templates

from jpcore.component import Component
from jpcore.justpy_config import config, AGGRID, AGGRID_ENTERPRISE,BOKEH,COOKIE_MAX_AGE, CRASH
from jpcore.justpy_config import DECKGL, FAVICON, HIGHCHARTS,HTML_404_PAGE,KATEX, LATENCY
from jpcore.justpy_config import NO_INTERNET, PLOTLY, SECRET_KEY, SESSION_COOKIE_NAME, SESSIONS
from jpcore.justpy_config import STATIC_DIRECTORY,STATIC_NAME
from jpcore.justpy_config import QUASAR, QUASAR_VERSION,TAILWIND, VEGA
from jpcore.justpy_config import FRONTEND_ENGINE_TYPE
from jpcore.template import Context
from jpcore.webpage import WebPage
from itsdangerous import Signer

# TODO refacto to object oriented version where this is a property of some instance of some class
cookie_signer = Signer(str(SECRET_KEY))

def create_component_file_list():
    file_list = []
    component_dir = os.path.join(STATIC_DIRECTORY, "components")
    if os.path.isdir(component_dir):
        for file in os.listdir(component_dir):
            if fnmatch.fnmatch(file, "*.js"):
                file_list.append(f"/components/{file}")
    return file_list


component_file_list = create_component_file_list()
grand_parent = pathlib.Path(__file__).parent.parent.resolve()
template_dir=f"{grand_parent}/justpy/templates"

lib_dir = os.path.join(template_dir, "js", FRONTEND_ENGINE_TYPE)
# remove .js extension
FRONTEND_ENGINE_LIBS = [fn[:-3]
                        for fn in os.listdir(lib_dir)
                        if fnmatch.fnmatch(fn, "*.js")
                        ]

TEMPLATES_DIRECTORY = config(
    "TEMPLATES_DIRECTORY", cast=str, default=template_dir
)

templates = Jinja2Templates(directory=TEMPLATES_DIRECTORY)

template_options = {
    "tailwind": TAILWIND,
    "quasar": QUASAR,
    "quasar_version": QUASAR_VERSION,
    "highcharts": HIGHCHARTS,
    "aggrid": AGGRID,
    "aggrid_enterprise": AGGRID_ENTERPRISE,
    "static_name": STATIC_NAME,
    "component_file_list": component_file_list,
    "no_internet": NO_INTERNET,
    "katex": KATEX,
    "plotly": PLOTLY,
    "bokeh": BOKEH,
    "deckgl": DECKGL,
    "vega": VEGA,
}

async def handle_event(data_dict, com_type=0, page_event=False):
    # com_type 0: websocket, con_type 1: ajax
    connection_type = {0: "websocket", 1: "ajax"}
    logging.info(
        "%s %s %s", "In event handler:", connection_type[com_type], str(data_dict)
    )
    event_data = data_dict["event_data"]
    try:
        p = WebPage.instances[event_data["page_id"]]
    except:
        logging.warning("No page to load")
        return
    event_data["page"] = p
    if com_type == 0:
        event_data["websocket"] = WebPage.sockets[event_data["page_id"]][
            event_data["websocket_id"]
        ]
    # The page_update event is generated by the reload_interval Ajax call
    if event_data["event_type"] == "page_update":
        build_list = p.build_list()
        return {"type": "page_update", "data": build_list}

    if page_event:
        c = p
    else:
        component_id = event_data["id"]
        c = Component.instances.get(component_id, None)
        if c is not None:
            event_data["target"] = c
        else:
            logging.warning(
                f"component with id {component_id} doesn't exist (anymore ...) it might have been deleted before the event handling was triggered"
            )

    try:
        if c is not None:
            before_result = await c.run_event_function("before", event_data, True)
    except:
        pass
    try:
        if c is not None:
            if hasattr(c, "on_" + event_data["event_type"]):
                event_result = await c.run_event_function(
                    event_data["event_type"], event_data, True
                )
            else:
                event_result = None
                logging.debug(f"{c} has no {event_data['event_type']} event handler")
        else:
            event_result = None
        logging.debug(f"Event result:{event_result}")
    except Exception as e:
        # raise Exception(e)
        if CRASH:
            print(traceback.format_exc())
            sys.exit(1)
        event_result = None
        # logging.info('%s %s', 'Event result:', '\u001b[47;1m\033[93mAttempting to run event handler:' + str(e) + '\033[0m')
        logging.info(
            "%s %s",
            "Event result:",
            "\u001b[47;1m\033[93mError in event handler:\033[0m",
        )
        logging.info("%s", traceback.format_exc())

    # If page is not to be updated, the event_function should return anything but None
    if event_result is None:
        if com_type == 0:  # WebSockets communication
            if LATENCY:
                await asyncio.sleep(LATENCY / 1000)
            await p.update()
        elif com_type == 1:  # Ajax communication
            build_list = p.build_list()
    try:
        if c is not None:
            after_result = await c.run_event_function("after", event_data, True)
    except:
        pass
    if com_type == 1 and event_result is None:
        dict_to_send = {
            "type": "page_update",
            "data": build_list,
            "page_options": {
                "display_url": p.display_url,
                "title": p.title,
                "redirect": p.redirect,
                "open": p.open,
                "favicon": p.favicon,
            },
        }
        return dict_to_send

# https://stackoverflow.com/questions/57412825/how-to-start-a-uvicorn-fastapi-in-background-when-testing-with-pytest
# https://github.com/encode/uvicorn/discussions/1103
# https://stackoverflow.com/questions/68603658/how-to-terminate-a-uvicorn-fastapi-application-cleanly-with-workers-2-when
class JustpyApp(Starlette):
    """
    a justpy application is a special Starlette application
    
      uses starlette Routing

    see
       https://www.starlette.io/routing/

       https://github.com/encode/starlette/blob/master/starlette/routing.py
    """
    # @Todo - legacy for SetRoute 
    app=None

    def __init__(self,**kwargs):
        # https://www.starlette.io/applications/
        Starlette.__init__(self,**kwargs)
        # @Todo - legacy for SetRoute 
        JustpyApp.app=self
    
    def route_as_text(self,route):
        """
        get a string representation of the given route
        """
        text= f"{route.__class__.__name__}(name: {route.name}, path: {route.path}, format: {route.path_format},  regex: {route.path_regex})"
        if isinstance(route,Route):
            text+=f"func: {route.endpoint.__name__}"
        return text
    
    def add_jproute(self,path:str,wpfunc:typing.Callable,name:str=None):
        """
        add a route for the given Webpage returning func
        
        Args:
            path(str): the path to use as route
            wpfunc(typing.Callable): a Webpage returning func
            name(str): the name of the route
        """
        endpoint=self.response(wpfunc)
        if name is None:
            name=wpfunc.__name__
        self.router.add_route(path,endpoint,name=name,include_in_schema=False)
    
    def jproute(self,
        path: str,
        name: typing.Optional[str] = None)-> typing.Callable:  # pragma: nocover
        """ 
        justpy route decorator
        
        function will we "wrapped" as a response and a route added
        
        Args:
            func(typing.Callable): the function to convert to a reponse
        """
        
        def routeResponse(func:typing.Callable)-> typing.Callable:
            """
            decorator for the given func
            
            Args:
                func(typing.Callable)
                
            Returns:
                Callable: an endpoint that has been routed
            
            """
            endpoint=self.response(func)
            self.router.add_route(
                path,
                endpoint,
                name=name if name is not None else func.__name__,
                include_in_schema=False,
            )
            self.route(path)
            return endpoint
        
        return routeResponse
    
    def response(self,func:typing.Callable):
        """
        response decorator converts a function to a response
        
        see also https://github.com/justpy-org/justpy/issues/532
        castAsEndPoint
        
        Args:
            func(typing.Callable): the function (returning a WebPage) to convert to a response
        """
        async def funcResponse(request)->HTMLResponse:
            """
            decorator function to apply the function to the request and
            return it as a response
            
            Args:
                request(Request): the request to apply the function to
                
            Returns:
                Response: a HTMLResponse applying the justpy infrastructure
            
            """
            new_cookie = self.handle_session_cookie(request)
            wp = await self.get_page_for_func(request, func)
            response = self.get_response_for_load_page(request, wp)
            response = self.set_cookie(request, response, wp, new_cookie)
            if LATENCY:
                await asyncio.sleep(LATENCY / 1000)
            return response
    
        # return the decorated function, thus allowing access to the func
        # parameter in the funcResponse later when applied 
        return funcResponse

    async def get_page_for_func(self, request, func:typing.Callable)->WebPage:
        """
        get the Webpage for the given func

        Args:
            request: the request to pass to the given function
            func: the function
            
        Returns:
            WebPage: the Webpage returned by the given function
        """
        # @TODO - get rid of the global func_to_run concept that isn't
        # in scope here (anymore) anyways
        func_to_run = func
        func_parameters = len(inspect.signature(func_to_run).parameters)
        assert (func_parameters < 2), f"Function {func_to_run.__name__} cannot have more than one parameter"
        if inspect.iscoroutinefunction(func_to_run):
            if func_parameters == 1:
                load_page = await func_to_run(request)
            else:
                load_page = await func_to_run()
        else:
            if func_parameters == 1:
                load_page = func_to_run(request)
            else:
                load_page = func_to_run()
        return load_page

    def get_response_for_load_page(self,request,load_page):
        """
        get the response for the given webpage
        
        Args:
            request(Request): the request to handle
            load_page(WebPage): the webpage to wrap with justpy and  
            return as a full HtmlResponse
        
        Returns:
            Reponse: the response for the given load_page
        """
        page_type = type(load_page)
        assert issubclass(
            page_type, WebPage
        ), f"Function did not return a web page but a {page_type.__name__}"
        if len(load_page) == 0 and not load_page.html:
            error_html="""<span style="color:red">Web page is empty - you might want to add components</span>"""
            return HTMLResponse(error_html, 500)
        page_options = {
            "reload_interval": load_page.reload_interval,
            "body_style": load_page.body_style,
            "body_classes": load_page.body_classes,
            "css": load_page.css,
            "head_html": load_page.head_html,
            "body_html": load_page.body_html,
            "display_url": load_page.display_url,
            "dark": load_page.dark,
            "title": load_page.title,
            "redirect": load_page.redirect,
            "highcharts_theme": load_page.highcharts_theme,
            "debug": load_page.debug,
            "events": load_page.events,
            "favicon": load_page.favicon if load_page.favicon else FAVICON,
        }
        if load_page.use_cache:
            page_dict = load_page.cache
        else:
            page_dict = load_page.build_list()
        template_options["tailwind"] = load_page.tailwind
        context = {
            "request": request,
            "page_id": load_page.page_id,
            "justpy_dict": json.dumps(page_dict, default=str),
            "use_websockets": json.dumps(WebPage.use_websockets),
            "options": template_options,
            "page_options": page_options,
            "html": load_page.html,
            "frontend_engine_type": FRONTEND_ENGINE_TYPE,
            "frontend_engine_libs": FRONTEND_ENGINE_LIBS
        }
        # wrap the context in a context object to make it available
        context_obj = Context(context)
        context["context_obj"] = context_obj
        response = templates.TemplateResponse(load_page.template_file, context)
        return response
    
    def handle_session_cookie(self,request) -> typing.Union[bool, Response]:
        """
        handle the session cookie for this request
        
        Returns:
            True if a new cookie and session has been created
        """
        # Handle web requests
        session_cookie = request.cookies.get(SESSION_COOKIE_NAME)
        new_cookie=None
        if SESSIONS:
            new_cookie = False
            if session_cookie:
                try:
                    session_id = cookie_signer.unsign(session_cookie).decode("utf-8")
                except:
                    return PlainTextResponse("Bad Session")
                request.state.session_id = session_id
                request.session_id = session_id
            else:
                # Create new session_id
                request.state.session_id = str(uuid.uuid4().hex)
                request.session_id = request.state.session_id
                new_cookie = True
                logging.debug(f"New session_id created: {request.session_id}")
        return new_cookie
    
    def set_cookie(self, request, response, load_page, new_cookie: typing.Union[bool, Response]):
        """
        set the cookie_value
        
        Args:
            request: the request 
            response: the response to be sent
            load_page(WebPage): the WebPage to handle
            new_cookie(bool|Response): True if there is a new cookie. Or Response if cookie was invalid
        """
        if isinstance(new_cookie, Response):
            return new_cookie
        if SESSIONS and new_cookie:
            cookie_value = cookie_signer.sign(request.state.session_id)
            cookie_value = cookie_value.decode("utf-8")
            response.set_cookie(
                SESSION_COOKIE_NAME, cookie_value, max_age=COOKIE_MAX_AGE, httponly=True
            )
            for k, v in load_page.cookies.items():
                response.set_cookie(k, v, max_age=COOKIE_MAX_AGE, httponly=True)
        return response

class JustpyAjaxEndpoint(HTTPEndpoint):
    """
    Justpy specific HTTPEndpoint/app (ASGI application)
    """
    
    def __init__(self,scope,receive,send):
        """ 
        constructor
        """
        HTTPEndpoint.__init__(self,scope, receive, send)

    async def post(self, request):
        """
        Handles post method. Used in Ajax mode for events when websockets disabled
        
        Args:
            request(Request): the request to handle
        """
        data_dict = await request.json()
        # {'type': 'event', 'event_data': {'event_type': 'beforeunload', 'page_id': 0}}
        if data_dict["event_data"]["event_type"] == "beforeunload":
            return await self.on_disconnect(data_dict["event_data"]["page_id"])

        session_cookie = request.cookies.get(SESSION_COOKIE_NAME)
        if SESSIONS and session_cookie:
            session_id = cookie_signer.unsign(session_cookie).decode("utf-8")
            data_dict["event_data"]["session_id"] = session_id

        # data_dict['event_data']['session'] = request.session
        msg_type = data_dict["type"]
        data_dict["event_data"]["msg_type"] = msg_type
        page_event = True if msg_type == "page_event" else False
        result = await handle_event(data_dict, com_type=1, page_event=page_event)
        if result:
            if LATENCY:
                await asyncio.sleep(LATENCY / 1000)
            return JSONResponse(result)
        else:
            return JSONResponse(False)

    async def on_disconnect(self, page_id):
        logging.debug(f"In disconnect Homepage")
        await WebPage.instances[
            page_id
        ].on_disconnect()  # Run the specific page disconnect function
        return JSONResponse(False)

class JustpyServer:
    """
    a justpy Server


    """

    def __init__(
        self,
        host: str = None,
        port: int = 10000,
        sleep_time: float = 0.5,
        mode: str = None,
        debug: bool = False,
    ):
        """
        constructor

        Args:
            port(int): the port
            host(str): the host
            sleep_time(float): the time to sleep after server process was started
            mode(str): None, direct or process. If None direct is used on MacOs and process on other platforms.
                process mode will run the task as a process and kill it with psutils, direct will use threading and
                trying shutdown with uvicorns built in shutdown method (as of 2022-09 this leads to error messages since
                the starlette router is not shutdown properly)
            debug(bool): if True switch debugging on
        """
        if host is None:
            host = socket.getfqdn()
            # host="127.0.0.1"
        self.host = host
        self.port = port
        self.sleep_time = sleep_time
        self.server = None
        self.proc = None
        self.thread = None
        if mode is None:
            if platform == "darwin":
                mode = "direct"
            else:
                mode = "process"
        self.mode = mode
        self.debug = debug
        self.running = False

    async def start(self, wpfunc,websockets: bool = True, **kwargs):
        """
        start a justpy server for the given webpage function wpfunc

        Args:
            wpfunc: the (async) function for the webpage


        """
        # this import is actually calling code ...
        import justpy as jp

        if self.mode == "direct":
            jp.justpy(
                wpfunc,
                host=self.host,
                port=self.port,
                start_server=False,
                websockets=websockets,
                kwargs=kwargs,
            )
            await asyncio.sleep(self.sleep_time)  # time for the server to start
            self.server = jp.get_server()
            self.thread = Thread(target=self.server.run)
            self.thread.start()
        elif self.mode == "process":
            needed_kwargs = {
                "host": self.host,
                "port": self.port,
                "start_server": True,
            }
            kwargs = {**needed_kwargs, **kwargs}
            self.proc = Process(
                target=jp.justpy,
                args=(wpfunc,),
                kwargs=kwargs,
            )
            self.proc.daemon = True
            self.proc.start()
        await asyncio.sleep(self.sleep_time)  # time for the server to start

    async def stop(self):
        """
        stop the server
        """
        # self.cancel()
        # https://stackoverflow.com/a/59089890/1497139
        # tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        # [task.cancel() for task in tasks]
        # await asyncio.gather(*tasks)
        # https://stackoverflow.com/questions/58133694/graceful-shutdown-of-uvicorn-starlette-app-with-websockets
        if self.server:
            # await asyncio.wait([jp.app.router.shutdown()],timeout=self.sleep_time)
            self.server.should_exit = True
            self.server.force_exit = True
            await asyncio.sleep(self.sleep_time)
            await self.server.shutdown()
        if self.thread:
            self.thread.join(timeout=self.sleep_time)
        if self.proc:
            pid = self.proc.pid
            parent = psutil.Process(pid)
            for child in parent.children(recursive=True):
                child.terminate()
            self.proc.terminate()

    def next_server(self):
        """
        get another similar server with the port number incremented by one
        """
        next_server = JustpyServer(
            port=self.port + 1,
            host=self.host,
            sleep_time=self.sleep_time,
            mode=self.mode,
            debug=self.debug,
        )
        return next_server

    def get_url(self, path):
        """
        get the url for the given path

        Args:
            path(str): the path
        Returns:
            str: the url for the path

        """
        url = f"http://{self.host}:{self.port}{path}"
        return url


# Justpy Tutorial demo show_demo from docs/tutorial/html_components.md
import justpy as jp

button_classes='m-2 bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded'

def show_demo():
    wp = jp.WebPage()
    b = jp.Button(text='Click to toggle show', a=wp, classes=button_classes)
    d = jp.Div(text='Toggled by show', classes='m-2 p-2 text-2xl border w-48', a=wp)
    b.d = d
    jp.Div(text='Will always show', classes='m-2', a=wp)

    def toggle_show(self, msg):
        self.d.show = not self.d.show

    b.on('click', toggle_show)

    b = jp.Button(text='Click to toggle visibility', a=wp, classes=button_classes)
    d = jp.Div(text='Toggled by visible', classes='m-2 p-2 text-2xl border w-48', a=wp)
    d.visibility_state = 'visible'
    b.d = d
    jp.Div(text='Will always show', classes='m-2', a=wp)

    def toggle_visible(self, msg):
        if self.d.visibility_state == 'visible':
            self.d.set_class('invisible')
            self.d.visibility_state = 'invisible'
        else:
            self.d.set_class('visible')
            self.d.visibility_state = 'visible'

    b.on('click', toggle_visible)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("show_demo",show_demo)



"""
Created on 2022-09-07

@author: hr
"""
import os
import unittest
from timeit import timeit

import hjson

from tests.basetest import Basetest
from addict import Dict

SLOW_TESTS = bool(os.environ.get('JP_SLOW_TESTS', False))


class TestJavaScriptObject(Basetest):
    """
    Tests JavaScript object conversion via hjson
    """

    # example options string see
    # https://www.highcharts.com/docs/getting-started/how-to-set-options
    options_string: str = """{
        chart: {
            renderTo: 'container',
            type: 'bar'
        },
        title: {
            text: 'Fruit Consumption'
        },
        xAxis: {
            categories: ['Apples', 'Bananas', 'Oranges']
        },
        yAxis: {
            title: {
                text: 'Fruit eaten'
            }
        },
        series: [{
            name: 'Jane',
            data: [1, 0, 4]
        }, {
            name: 'John',
            data: [5, 7, 3]
        }]
    }"""

    def test_decode(self):
        """
        test JavaScript object decoding
        """
        options = Dict(hjson.loads(self.options_string.encode("ascii", "ignore")))
        debug = self.debug
        # debug=True
        if debug:
            print(options)
        self.assertTrue("chart" in options)
        pass

    def test_decode_no_separating_spaces(self):
        """
        test JavaScript object decoding
        """
        # example options string see
        # https://www.highcharts.com/docs/getting-started/how-to-set-options
        options_string = """{
    title:{
        text:'Fruit Consumption'
    },
    xAxis: {
        categories: ['Apples', 'Bananas', 'Oranges']
    },
    series:[]
}"""
        options = Dict(hjson.loads(options_string.encode("ascii", "ignore")))
        debug = self.debug
        # debug=True
        if debug:
            print(options)
        self.assertTrue("series" in options)
        self.assertTrue("title" in options)
        self.assertTrue("text" in options["title"])

    def test_decode_dirty_js(self):
        """
        test decoding of dirty JavaScript objects
        """
        dirty_js = """
             {
          key: "value",
          "key2":"value"
        }
        """
        options = hjson.loads(dirty_js)
        self.assertTrue("key2" in options)
        self.assertTrue("key" in options)

    def test_decode_dirty_js2(self):
        """
        test decoding of dirty JavaScript objects
        cf. https://stackoverflow.com/questions/34812821/bad-json-keys-are-not-quoted
        """
        dirty_js = """
             {
          key:"value",
          "key2":"value"
        }
        """
        options = hjson.loads(dirty_js)
        self.assertTrue("key2" in options)
        self.assertTrue("key" in options)

    @unittest.skipIf(not SLOW_TESTS, "JP_SLOW_TESTS")
    def test_decode_time(self):
        """
        test decoding time
        """
        no_runs = 1000
        elapsed = timeit(lambda: hjson.loads(self.options_string.encode("ascii", "ignore")),
                         number=no_runs)
        print(f'Time elapsed for {no_runs} runs: {elapsed:.2f}s')
        self.assertGreater(elapsed, 0.0)



'''
Created on 2022-09-08

@author: wf
'''
import asyncio

from selenium.webdriver.remote.webdriver import BaseWebDriver
from selenium.webdriver.support.wait import WebDriverWait

from tests.base_server_test import BaseAsynctest
from tests.browser_test import SeleniumBrowsers
from tests.basetest import Basetest
from examples.basedemo import Demo

class BaseSeleniumTest(BaseAsynctest):
    """
    Base class for Selenium tests
    """
    
    async def setUp(
        self, 
        port:int=8123, 
        host:str="127.0.0.1", 
        sleep_time=None, 
        with_server=True, 
        debug=False, 
        profile=True, 
        mode=None):
        await BaseAsynctest.setUp(self, port=port, host=host, sleep_time=sleep_time, with_server=with_server, debug=debug, profile=profile, mode=mode)
        await asyncio.sleep(self.server.sleep_time)
        self.browser = SeleniumBrowsers(headless=Basetest.inPublicCI()).getFirst()
        
    async def getBrowserForDemo(self):
        """
        get the browser for a Demo test
        """
        browser=SeleniumBrowsers(headless=Basetest.inPublicCI()).getFirst()
        await asyncio.sleep(self.server.sleep_time)
        Demo.testmode = True
        return browser

    def get_waiting_browser(self, browser: BaseWebDriver, timeout: float = 5.0):
        """
        Return a waiting webdriver for the given driver/browser

        Args:
            browser: WebDriver
            timeout: maximum time to wait
        """
        driver = WebDriverWait(browser, timeout)
        return driver



# Justpy Tutorial demo commands_demo2 from docs/tutorial/working_with_html.md
import justpy as jp

def commands_demo2():
    wp = jp.WebPage()
    root = jp.Div(a=wp)
    c1 = jp.Div(a=root)
    c2 = jp.P(classes='m-2 p-2 text-red-500 text-xl', a=c1, text='Paragraph 1')
    c3 = jp.P(classes='m-2 p-2 text-blue-500 text-xl', a=c1, text='Paragraph 2')
    c4 = jp.P(classes='m-2 p-2 text-green-500 text-xl', a=c1, text='Paragraph 3')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("commands_demo2",commands_demo2)



"""
Created on 2022-08-24

@author: em
https://justpy.io/tutorial/custom_components/#tab-group-component
"""
import justpy as jp
from justpy import Div
from examples.tutorial.tab_group_component import Tabs, TabsPills

my_chart_def = """
{
        chart: {
            type: 'bar'
        },
        title: {
            text: 'Fruit Consumption'
        },
        xAxis: {
            categories: ['Apples', 'Bananas', 'Oranges']
        },
        yAxis: {
            title: {
                text: 'Fruit eaten'
            }
        },
        series: [{
            name: 'Jane',
            data: [1, 0, 4],
            animation: false
        }, {
            name: 'John',
            data: [5, 7, 3],
            animation: false
        }]
}
"""
# https://dog.ceo/api/breed/papillon/images/random
pics_french_bulldogs = ["5458", "7806", "5667", "4860"]
pics_papillons = ["5037", "2556", "7606", "8241"]


def tab_change(self, msg):
    print("in change", msg)


def tab_comp_test():
    wp = jp.WebPage(data={"tab": "id2556"})

    t = Tabs(a=wp, classes="w-3/4 m-4", style="", animation=True, content_height=550)
    for chart_type in ["bar", "column", "line", "spline"]:
        d = jp.Div(style=Tabs.wrapper_style, delete_flag=True)
        my_chart = jp.HighCharts(
            a=d,
            classes="m-2 p-2 border",
            style="width: 1000px;",
            options=my_chart_def,
            use_cache=False,
        )
        my_chart.options.chart.type = chart_type
        my_chart.options.title.text = f"Chart of Type {chart_type.capitalize()}"
        my_chart.options.subtitle.text = f"Subtitle {chart_type.capitalize()}"
        t.add_tab(f"id{chart_type}", f"{chart_type}", d)

    d_flex = Div(classes="flex", a=wp)  # Container for the two dog pictures tabs

    t = Tabs(
        a=d_flex,
        classes=" w-1/2 m-4",
        animation=True,
        content_height=550,
        model=[wp, "tab"],
        change=tab_change,
    )
    for pic_id in pics_papillons:
        d = jp.Div(style=Tabs.wrapper_style)
        jp.Img(
            src=f"https://images.dog.ceo/breeds/papillon/n02086910_{pic_id}.jpg", a=d
        )
        t.add_tab(f"id{pic_id}", f"Pic {pic_id}", d)

    t = TabsPills(
        a=d_flex,
        classes="w-1/2 m-4",
        animation=True,
        content_height=550,
        change=tab_change,
    )
    for pic_id in pics_french_bulldogs:
        d = jp.Div(style=Tabs.wrapper_style)
        jp.Img(
            src=f"https://images.dog.ceo/breeds/bulldog-french/n02108915_{pic_id}.jpg",
            a=d,
        )
        t.add_tab(f"id{pic_id}", f"Pic {pic_id}", d)

    input_classes = "w-1/3 m-2 bg-gray-200 border-2 border-gray-200 rounded w-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"

    in1 = jp.Input(classes=input_classes, model=[wp, "tab"], a=wp)

    return wp


from examples.basedemo import Demo

Demo("tabgroup test", tab_comp_test)



# Justpy Tutorial demo product_app_c from docs/blog/vue_comparison.md
import justpy as jp

products = [
            {'id': 1, 'quantity': 1, 'name': 'Compass'},
            {'id': 2, 'quantity': 0, 'name': 'Jacket'},
            {'id': 3, 'quantity': 5, 'name': 'Hiking Socks'},
            {'id': 4, 'quantity': 2, 'name': 'Suntan Lotion'},
            ]


class Products(jp.Div):

    def __init__(self, **kwargs):
        self.products = products
        super().__init__(**kwargs)
        self.ul = jp.Ul(a=self)
        self.total_inventory = jp.H2(text='Totals', a=self)

    def total_products(self):
        total = 0
        for product in self.products:
            total += product["quantity"]
        return total

    def react2(self, data):
        self.ul.delete_components()
        for product in self.products:
            item = jp.Li(a=self.ul)
            jp.Input(type='number', product=product, a=item, value=product["quantity"], input='self.product["quantity"] = self.value')
            jp.Span(text=f'{product["quantity"]} {product["name"]}', a=item, style='margin-left: 10px')
            if product["quantity"] == 0:
                jp.Span(text=' - OUT OF STOCK', a=item)
            jp.Button(text='Add', a=item, product=product, click='self.product["quantity"] += 1', style='margin-left: 10px')
        self.total_inventory.text = f'Total Inventory: {self.total_products()}'


def product_app_c():
    wp = jp.WebPage(tailwind=False)
    Products(a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("product_app_c",product_app_c)



from types import MethodType
from addict import Dict
import json, copy, inspect, sys, re
from html.parser import HTMLParser, tagfind_tolerant, attrfind_tolerant
from html.entities import name2codepoint
from html import unescape
from jinja2 import Template
import asyncio
from jpcore.tailwind import Tailwind
import logging
import httpx
from jpcore.template import PageOptions
from jpcore.component import Component
from jpcore.webpage import WebPage as BaseWebPage

# Dictionary for translating from tag to class
_tag_class_dict = {}


def parse_dict(cls):
    """
    Decorator for component class definitions that updates _tag_class_dict so that the parser can recognize new components
    Required only for components not defined in this module
    """
    _tag_class_dict[cls.html_tag] = cls
    return cls


class JustPy:
    loop = None
    LOGGING_LEVEL = logging.DEBUG
    component_registry = {}
    
class WebPage(BaseWebPage):
    """
    enhanced WebPage as per documentation
    """
    
    def write(self, md_text):
        """
        write to a Markdown div
        """
        assert _has_markdown, "Markdown not installed"
        return Markdown(
            markdown=md_text, a=self, classes="m-2 p-2", all_extensions=True
        )

    def equation(self, eq_text):
        """
        return an Equation linked to this webpage with the given equation text
        """
        return Equation(equation=eq_text, a=self, classes="m-2 p-2")

def register_component(component_class, tag, attributes=[]):
    JustPy.component_registry[tag] = {
        "class": component_class,
        "attributes": attributes,
    }


class Register:
    def __init__(self, tag, attributes=[], **kwargs):
        self.tag = tag
        self.kwargs = kwargs
        self.attributes = attributes

    def __call__(self, cls, **kwargs):
        register_component(cls, self.tag, self.attributes)
        return cls

class TailwindVersion1Page(WebPage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tailwind = False
        self.head_html = """
        <link rel="stylesheet" href="https://unpkg.com/tailwindcss@^1.5/dist/base.min.css" />
        <link rel="stylesheet" href="https://unpkg.com/tailwindcss@^1.5/dist/components.min.css" />
        <link rel="stylesheet" href="https://unpkg.com/@tailwindcss/typography@0.2.x/dist/typography.min.css"/>
        <link rel="stylesheet" href="https://unpkg.com/tailwindcss@^1.5/dist/utilities.min.css" />
        """

class TailwindUIPage(WebPage):
    # https://tailwindui.com/components

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.template_file = "tailwindui.html"

class JustpyBaseComponent(Component):
    """
    the base class for all Justpy components
    """
    temp_flag = True
    delete_flag = True
    needs_deletion = False

    def __init__(self, **kwargs):
        """
        constructor
        """
        cls = JustpyBaseComponent
        temp = kwargs.get("temp", cls.temp_flag)
        delete_flag = kwargs.get("delete_flag", cls.delete_flag)
        if temp and delete_flag:
            self.id = None
        else:
            self.id = cls.next_id
            cls.next_id += 1
        self.events = []
        self.event_modifiers = Dict()
        self.transition = None
        self.allowed_events = []

    def initialize(self, **kwargs):
        for k, v in kwargs.items():
            self.__setattr__(k, v)
        self.set_keyword_events(**kwargs)
        for com in ["a", "add_to"]:
            if com in kwargs.keys():
                kwargs[com].add_component(self)

    def set_keyword_events(self, **kwargs):
        for e in self.allowed_events:
            for prefix in ["", "on", "on_"]:
                if prefix + e in kwargs.keys():
                    cls = JustpyBaseComponent
                    if not self.id:
                        self.id = cls.next_id
                        cls.next_id += 1
                    fn = kwargs[prefix + e]
                    if isinstance(fn, str):
                        fn_string = f"def oneliner{self.id}(self, msg):\n {fn}"
                        exec(fn_string)
                        self.on(e, locals()[f"oneliner{self.id}"])
                    else:
                        self.on(e, fn)
                    break

    def delete(self):
        if self.needs_deletion:
            if self.delete_flag:
                JustpyBaseComponent.instances.pop(self.id, None)
                self.needs_deletion = False

    def on(
        self,
        event_type,
        func,
        *,
        debounce=None,
        throttle=None,
        immediate=False,
    ):
        if event_type in self.allowed_events:
            cls = JustpyBaseComponent
            if not self.id:
                self.id = cls.next_id
                cls.next_id += 1
            cls.instances[self.id] = self
            self.needs_deletion = True
            if inspect.ismethod(func):
                setattr(self, "on_" + event_type, func)
            else:
                setattr(self, "on_" + event_type, MethodType(func, self))
            if event_type not in self.events:
                self.events.append(event_type)
            if debounce:
                self.event_modifiers[event_type].debounce = {
                    "value": debounce,
                    "timeout": None,
                    "immediate": immediate,
                }
            elif throttle:
                self.event_modifiers[event_type].throttle = {
                    "value": throttle,
                    "timeout": None,
                }
        else:
            raise Exception(f"No event of type {event_type} supported")

    def remove_event(self, event_type):
        if event_type in self.events:
            self.events.remove(event_type)

    def has_event_function(self, event_type):
        if getattr(self, "on_" + event_type, None):
            return True
        else:
            return False

    def has_class(self, class_name):
        return class_name in self.classes.split()

    def remove_class(self, tw_class):
        class_list = self.classes.split()
        try:
            class_list.remove(tw_class)
        except:
            pass
        self.classes = " ".join(class_list)

    def hidden(self, flag=True):
        if flag:
            self.set_class("hidden")
        else:
            self.remove_class("hidden")

    def hidden_toggle(self):
        if self.has_class("hidden"):
            self.remove_class("hidden")
        else:
            self.set_class("hidden")

    async def update(self, socket=None, *, react=None):
        if react:
            self.react([])
        component_dict = self.convert_object_to_dict()
        if socket:
            await socket.send_json({"type": "component_update", "data": component_dict})
        else:
            pages_to_update = list(self.pages.values())
            for page in pages_to_update:
                try:
                    websocket_dict = WebPage.sockets[page.page_id]
                except:
                    continue
                for websocket in list(websocket_dict.values()):
                    try:
                        # WebPage.loop.create_task(websocket.send_json({'type': 'component_update', 'data': component_dict}))
                        await websocket.send_json(
                            {"type": "component_update", "data": component_dict}
                        )
                    except:
                        print("Problem with websocket in component update, ignoring")
        return self

    def check_transition(self):
        if self.transition and (not self.id):
            cls = JustpyBaseComponent
            self.id = cls.next_id
            cls.next_id += 1

    async def run_method(self, command, websocket):
        await websocket.send_json(
            {"type": "run_method", "data": command, "id": self.id}
        )
        # So the page itself does not update, return True not None
        return True

    def remove_page_from_pages(self, wp: WebPage):
        self.pages.pop(wp.page_id)

    def add_page(self, wp: WebPage):
        self.pages[wp.page_id] = wp

    def add_page_to_pages(self, wp: WebPage):
        self.pages[wp.page_id] = wp

    def set_model(self, value):
        if hasattr(self, "model"):
            if len(self.model) == 2:
                self.model[0].data[self.model[1]] = value
            else:
                self.model[0][self.model[1]] = value

    def get_model(self):
        if len(self.model) == 2:
            model_value = self.model[0].data[self.model[1]]
        else:
            model_value = self.model[0][self.model[1]]
        return model_value

    async def run_event_function(
        self, event_type, event_data, create_namespace_flag=True
    ):
        event_function = getattr(self, "on_" + event_type)
        if create_namespace_flag:
            function_data = Dict(event_data)
        else:
            function_data = event_data
        if inspect.iscoroutinefunction(event_function):
            event_result = await event_function(function_data)
        else:
            event_result = event_function(function_data)
        return event_result

    @staticmethod
    def convert_dict_to_object(d):
        obj = globals()[d["class_name"]]()
        for obj_prop in d["object_props"]:
            obj.add(JustpyBaseComponent.convert_dict_to_object(obj_prop))
        for k, v in d.items():
            obj.__dict__[k] = v
        for k, v in d["attrs"].items():
            obj.__dict__[k] = v
        return obj


class HTMLBaseComponent(JustpyBaseComponent):
    """
    Base Component for all HTML components
    """

    attributes = []
    html_tag = "div"
    vue_type = "html_component"  # Vue.js component name

    html_global_attributes = [
        "accesskey",
        "class",
        "contenteditable",
        "dir",
        "draggable",
        "dropzone",
        "hidden",
        "id",
        "lang",
        "spellcheck",
        "style",
        "tabindex",
        "title",
    ]

    attribute_list = [
        "id",
        "vue_type",
        "show",
        "events",
        "event_modifiers",
        "classes",
        "style",
        "set_focus",
        "html_tag",
        "class_name",
        "event_propagation",
        "inner_html",
        "animation",
        "debug",
        "transition",
    ]

    # not_used_global_attributes = ['dropzone', 'translate', 'autocapitalize',
    #                               'itemid', 'itemprop', 'itemref', 'itemscope', 'itemtype']

    # Additions to global attributes to add to attrs dict apart from id and style.
    used_global_attributes = [
        "contenteditable",
        "dir",
        "tabindex",
        "title",
        "accesskey",
        "draggable",
        "lang",
        "spellcheck",
    ]

    # https://developer.mozilla.org/en-US/docs/Web/HTML/Element

    # windows_events = ['afterprint', 'beforeprint', 'beforeunload', 'error', 'hashchange', 'load',
    #                   'message', 'offline', 'online', 'pagehide', 'pageshow', 'popstate',
    #                   'resize', 'storage', 'unload']
    # form_events = ['blur', 'change', 'contextmenu', 'focus', 'input', 'invalid', 'reset', 'search', 'select', 'submit']
    # keyboard_events = ['keydown', 'keypress', 'keyup']
    # mouse_events = ['click', 'dblclick', 'mousedown', 'mousemove', 'mouseout', 'mouseover', 'mouseup', 'wheel',
    #                 'mouseenter', 'mouseleave']
    # allowed_events = ['click', 'mouseover', 'mouseout', 'mouseenter', 'mouseleave', 'input', 'change',
    #                   'after', 'before', 'keydown', 'keyup', 'keypress', 'focus', 'blur']

    # allowed_events = ['click', 'mouseover', 'mouseout', 'mouseenter', 'mouseleave', 'input', 'change',
    #                   'after', 'before', 'keydown', 'keyup', 'keypress', 'focus', 'blur', 'submit',
    #                   'dragstart', 'dragover', 'drop']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_name = type(self).__name__
        self.debug = False
        self.inner_html = ""
        self.animation = False
        self.pages = (
            {}
        )  # Dictionary of pages the component is on. Not managed by framework.
        self.show = True
        self.set_focus = False
        self.classes = ""
        self.slot = None
        self.scoped_slots = {}  # For Quasar and other Vue.js based components
        self.style = ""
        self.directives = []
        self.data = {}
        self.drag_options = None
        self.allowed_events = [
            "click",
            "mouseover",
            "mouseout",
            "mouseenter",
            "mouseleave",
            "input",
            "change",
            "after",
            "before",
            "keydown",
            "keyup",
            "keypress",
            "focus",
            "blur",
            "submit",
            "dragstart",
            "dragover",
            "drop",
            "click__out",
        ]
        self.events = []
        self.event_modifiers = Dict()
        self.additional_properties = (
            []
        )  # Additional fields to get from the JavasScript event object
        self.event_propagation = True  # If True events are propagated
        self.prop_list = []  # For components from libraries like quasar

        self.initialize(**kwargs)

    def __len__(self):
        if hasattr(self, "components"):
            return len(self.components)
        else:
            return 0

    def __repr__(self):
        name = self.name if hasattr(self, "name") else "No name"
        return f"{self.__class__.__name__}(id: {self.id}, html_tag: {self.html_tag}, vue_type: {self.vue_type}, name: {name}, number of components: {len(self)})"

    def add_to_page(self, wp: WebPage):
        wp.add_component(self)

    def add_to(self, *args):
        for c in args:
            c.add_component(self)

    def add_attribute(self, attr, value):
        self.attrs[attr] = value

    def add_event(self, event_type):
        if event_type not in self.allowed_events:
            self.allowed_events.append(event_type)

    def add_allowed_event(self, event_type):
        self.add_event(event_type)

    def add_scoped_slot(self, slot, c):
        self.scoped_slots[slot] = c

    def to_html(self, indent=0, indent_step=0, format=True):
        block_indent = " " * indent
        if format:
            ws = "\n"
        else:
            ws = ""
        s = f"{block_indent}<{self.html_tag} "
        d = self.convert_object_to_dict()
        for attr, value in d["attrs"].items():
            if value:
                s = f'{s}{attr}="{value}" '
        if self.classes:
            s = f'{s}class="{self.classes}"/>{ws}'
        else:
            s = f"{s}/>{ws}"
        return s

    def react(self, data):
        return

    def convert_object_to_dict(self):
        d = {}
        # Add id if CSS transition is defined
        if self.transition:
            self.check_transition()
        if self.id:
            d["attrs"] = {"id": str(self.id)}
        else:
            d["attrs"] = {}
        for attr in HTMLBaseComponent.attribute_list:
            d[attr] = getattr(self, attr)
        d["directives"] = {}
        for i in self.directives:
            if i[0:2] == "v-":  # It is a directive
                try:
                    d["directives"][i[2:]] = getattr(self, i.replace("-", "_"))
                except:
                    pass
        for i in (
            self.prop_list + self.attributes + HTMLBaseComponent.used_global_attributes
        ):
            try:
                d["attrs"][i] = getattr(self, i)
            except:
                pass
            if i in ["in", "from"]:  # Attributes that are also python reserved words
                try:
                    d["attrs"][i] = getattr(self, "_" + i)
                except:
                    pass
            if "-" in i:
                s = i.replace("-", "_")  # kebab case to snake case
                try:
                    d["attrs"][i] = getattr(self, s)
                except:
                    pass
        # Name is a special case. Allow it to be defined for all
        try:
            d["attrs"]["name"] = self.name
        except:
            pass
        d["scoped_slots"] = {}
        for s in self.scoped_slots:
            d["scoped_slots"][s] = self.scoped_slots[s].convert_object_to_dict()
        if self.additional_properties:
            d["additional_properties"] = self.additional_properties
        if self.drag_options:
            d["drag_options"] = self.drag_options
        return d


class Div(HTMLBaseComponent):
    """ 
    A general purpose html container
    This is a component that other components can be added to
    """

    html_tag = "div"

    def __init__(self, **kwargs):
        """
        constructor
        """
        self.html_entity = False
        self.children = []
        super().__init__(**kwargs)
        self.components = self.children.copy()

    def delete(self):
        if self.delete_flag:
            for c in self.components:
                c.delete()
            if self.needs_deletion:
                JustpyBaseComponent.instances.pop(self.id, None)
            self.components = []

    def __getitem__(self, index):
        return self.components[index]

    def add_component(self, child, position=None, slot=None):
        """
        add a component
        
        Args:
            child: the component to add
            position: the position to add to (append if None) 
            slot: if given set the slot of the child
        """
        if slot:
            child.slot = slot
        if position is None:
            self.components.append(child)
        else:
            self.components.insert(position, child)
        return self

    def delete_components(self):
        for c in self.components:
            c.delete()
        self.components = []

    def add(self, *args):
        for component in args:
            self.add_component(component)
        return self

    def __add__(self, child):
        self.add_component(child)
        return self

    def __iadd__(self, child):
        self.add_component(child)
        return self

    def add_first(self, child):
        self.add_component(child, 0)

    def remove_component(self, component):
        try:
            self.components.remove(component)
        except:
            raise Exception(
                "Component cannot be removed because it is not contained in element"
            )
        return self

    def remove(self, component):
        self.remove_component(component)

    def get_components(self):
        return self.components

    def first(self):
        return self.components[0]

    def last(self):
        return self.components[-1]

    def to_html(self, indent=0, indent_step=0, format=True):
        block_indent = " " * indent
        if format:
            ws = "\n"
        else:
            ws = ""
        s = f"{block_indent}<{self.html_tag} "
        d = self.convert_object_to_dict()
        for attr, value in d["attrs"].items():
            if value:
                s = f'{s}{attr}="{value}" '
        if self.style:
            s = f'{s}style="{self.style}"'
        if self.classes:
            s = f'{s}class="{self.classes}">{ws}'
        else:
            s = f"{s}>{ws}"
        if self.inner_html:
            s = f"{s}{self.inner_html}</{self.html_tag}>{ws}"
            return s
        try:
            s = f"{s}{self.text}{ws}"
        except:
            pass
        for c in self.components:
            s = f"{s}{c.to_html(indent + indent_step, indent_step, format)}"
        s = f"{s}{block_indent}</{self.html_tag}>{ws}"
        return s

    def model_update(self):
        # [wp, 'text'] for example
        # self.text = str(self.model[0].data[self.model[1]])
        self.text = self.get_model()

    def build_list(self):
        object_list = []
        for i, obj in enumerate(self.components):
            obj.react(self.data)
            d = obj.convert_object_to_dict()
            object_list.append(d)
        return object_list

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        if hasattr(self, "model"):
            self.model_update()
        d["object_props"] = self.build_list()
        if hasattr(self, "text"):
            self.text = str(self.text)
            d["text"] = self.text
            # Handle HTML entities. Warning: They should be in their own span or div. Setting inner_html overrides all else in container
            if self.html_entity:
                d["inner_html"] = self.text
        return d


class Input(Div):
    # Edge and Internet explorer do not support the input event for checkboxes and radio buttons. Use change instead
    # IMPORTANT: Scope of name of radio buttons is the whole page and not the form unless form is specified

    html_tag = "input"
    attributes = [
        "accept",
        "alt",
        "autocomplete",
        "autofocus",
        "checked",
        "dirname",
        "disabled",
        "form",
        "formaction",
        "formenctype",
        "formmethod",
        "formnovalidate",
        "formtarget",
        "height",
        "list",
        "max",
        "maxlength",
        "min",
        "minlength",
        "multiple",
        "name",
        "pattern",
        "placeholder",
        "readonly",
        "required",
        "size",
        "src",
        "step",
        "type",
        "value",
        "width",
    ]

    def __init__(self, **kwargs):

        self.value = ""
        self.checked = False
        self.debounce = 200  # 200 millisecond default debounce for events
        self.no_events = False
        # Types for input element:
        # ['button', 'checkbox', 'color', 'date', 'datetime-local', 'email', 'file', 'hidden', 'image',
        # 'month', 'number', 'password', 'radio', 'range', 'reset', 'search', 'submit', 'tel', 'text', 'time', 'url', 'week']
        self.type = "text"
        self.form = None
        super().__init__(**kwargs)

        def default_input(self, msg):
            return self.before_event_handler(msg)

        if not self.no_events:
            self.on("before", default_input)

    def __repr__(self):
        num_components = len(self.components)
        return f"{self.__class__.__name__}(id: {self.id}, html_tag: {self.html_tag}, input_type: {self.type}, vue_type: {self.vue_type}, value: {self.value}, checked: {self.checked}, number of components: {num_components})"

    def before_event_handler(self, msg):
        logging.debug(
            "%s %s %s %s %s", "before ", self.type, msg.event_type, msg.input_type, msg
        )
        if msg.event_type not in ["input", "change", "select"]:
            return
        if msg.input_type == "checkbox":
            # The checked field is boolean
            self.checked = msg.checked
            if hasattr(self, "model"):
                self.model[0].data[self.model[1]] = msg.checked
        elif msg.input_type == "radio":
            # If a radio field, all other radio buttons with same name need to have value changed
            # If form is specified, the scope is that form. If not, it is the whole page
            self.checked = True
            if self.form:
                Input.radio_button_set(self, self.form)
            else:
                Input.radio_button_set(self, msg.page)
            if hasattr(self, "model"):
                self.model[0].data[self.model[1]] = msg.value
            self.value = msg.value
        else:
            if msg.input_type == "number":
                try:
                    msg.value = int(msg.value)
                except:
                    msg.value = float(msg.value)
            if hasattr(self, "model"):
                # self.model[0].data[self.model[1]] = msg.value
                self.set_model(msg.value)
            self.value = msg.value

    @staticmethod
    def radio_button_set(radio_button, container):
        # Set all radio buttons in container with same name as radio_button to unchecked
        if hasattr(container, "components"):
            for c in container.components:
                if hasattr(c, "name"):
                    if c.name == radio_button.name and not radio_button.id == c.id:
                        c.checked = False
                Input.radio_button_set(radio_button, c)

    @staticmethod
    def radio_button_set_model_update(radio_button, container, model_value):
        for c in container.components:
            if hasattr(c, "name"):
                if c.name == radio_button.name:
                    if c.value == model_value:
                        c.checked = True
                    else:
                        c.checked = False
            Input.radio_button_set_model_update(radio_button, c, model_value)

    def model_update(self):
        # update_value = self.model[0].data[self.model[1]]
        update_value = self.get_model()
        if self.type == "checkbox":
            self.checked = update_value
        elif self.type == "radio":
            model_value = update_value
            if self.form:
                Input.radio_button_set_model_update(self, self.form, model_value)
            else:
                Input.radio_button_set_model_update(self, self.model[0], model_value)
        else:
            self.value = update_value

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["debounce"] = self.debounce
        d[
            "input_type"
        ] = self.type  # Needed for vue component updated life hook and event handler
        if self.type in ["text", "textarea"]:
            d["value"] = str(self.value)
        else:
            d["value"] = self.value
        d["attrs"]["value"] = self.value
        d["checked"] = self.checked
        if not self.no_events:
            if self.type in ["radio", "checkbox", "select"] or self.type == "file":
                # Not all browsers create input event
                if "change" not in self.events:
                    self.events.append("change")
            else:
                if "input" not in self.events:
                    self.events.append("input")
        if self.checked:
            d["attrs"]["checked"] = True
        else:
            d["attrs"]["checked"] = False
        try:
            d["attrs"]["form"] = self.form.id
        except:
            pass

        return d


class InputChangeOnly(Input):
    """
    Does not generate the 'input' event. Generates the 'change' event. Leaves other events unchanged.
    Use if you don't need to look at each character typed. Saves interaction with the server
    The 'change' event docs:
    https://developer.mozilla.org/en-US/docs/Web/API/HTMLElement/change_event
    Salient: When the element loses focus after its value was changed, but not committed (e.g., after editing the value
    of <textarea> or <input type="text">) or when Enter is pressed.
    """

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        if "input" in d["events"]:
            d["events"].remove("input")
        if "change" not in d["events"]:
            d["events"].append("change")
        return d


class Form(Div):

    html_tag = "form"
    attributes = [
        "accept-charset",
        "action",
        "autocomplete",
        "enctype",
        "method",
        "name",
        "novalidate",
        "target",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        def default_submit(self, msg):
            print("Default form submit", msg.form_data)
            return True

        if not self.has_event_function("submit"):
            # If an event handler is not  assigned, the front end cannot stop the default page request that happens when a form is submitted
            self.on("submit", default_submit)


class Label(Div):

    html_tag = "label"
    attributes = [
        "for",
        "form",
    ]  # In JustPy these are components, not ids of component like in HTML

    def __init__(self, **kwargs):
        self.for_component = None
        super().__init__(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        try:
            d["attrs"]["for"] = self.for_component.id
        except:
            pass
        try:
            d["attrs"]["form"] = self.form.id
        except:
            pass
        return d


class Textarea(Input):

    html_tag = "textarea"
    attributes = [
        "autofocus",
        "cols",
        "dirname",
        "disabled",
        "form",
        "maxlength",
        "name",
        "placeholder",
        "readonly",
        "required",
        "rows",
        "wrap",
        "value",
    ]

    def __init__(self, **kwargs):
        self.rows = "4"
        self.cols = "50"
        super().__init__(**kwargs)
        self.type = "textarea"
        self.input_type = "text"


class Select(Input):
    # Need to set value of select on creation, otherwise blank line will show on page update
    html_tag = "select"
    attributes = [
        "autofocus",
        "disabled",
        "form",
        "multiple",
        "name",
        "required",
        "size",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "select"


class A(Div):

    html_tag = "a"
    attributes = [
        "download",
        "href",
        "hreflang",
        "media",
        "ping",
        "rel",
        "target",
        "type",
    ]

    def __init__(self, **kwargs):

        self.href = None
        self.bookmark = None  # The component on page to jump to or scroll to
        self.title = ""
        self.rel = "noopener noreferrer"
        self.download = None  # If attribute is set, file is downloaded, only works html 5  https://www.w3schools.com/tags/att_a_download.asp
        self.target = "_self"  # _blank, _self, _parent, _top, framename
        # https://developer.mozilla.org/en-US/docs/Web/API/Element/scrollIntoView
        self.scroll = False  # If True, scrolling is enabled
        self.scroll_option = "smooth"  # One of "auto" or "smooth".
        self.block_option = "start"  # One of "start", "center", "end", or "nearest". Defaults to "start".
        self.inline_option = "nearest"  # One of "start", "center", "end", or "nearest". Defaults to "nearest".
        super().__init__(**kwargs)

        if not kwargs.get("click"):

            def default_click(self, msg):
                return True

            self.on("click", default_click)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["scroll"] = self.scroll
        d["scroll_option"] = self.scroll_option
        d["block_option"] = self.block_option
        d["inline_option"] = self.inline_option
        if self.bookmark is not None:
            self.href = "#" + str(self.bookmark.id)
            self.scroll_to = str(self.bookmark.id)
        if d["scroll"]:
            d["scroll_to"] = self.scroll_to
        d["attrs"]["href"] = self.href
        d["attrs"]["target"] = self.target
        if self.download is not None:
            d["attrs"]["download"] = self.download
        return d


Link = A  # The 'Link' name is more descriptive and can be used instead


class Icon(Div):
    def __init__(self, **kwargs):
        self.icon = "dog"  # Default icon
        super().__init__(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["classes"] = self.classes + " fa fa-" + self.icon
        return d


class EditorMD(Textarea):
    # https://www.cssportal.com/style-input-range/   style an input range
    # Set the page's tailwind attribute to False for preview to work
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.debounce = 0
        self.input_type = "textarea"
        self.vue_type = "editorjp"
        self.html_tag = "textarea"


class Space(Div):

    # Creates a span with hard spaces.

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.num = kwargs.get("num", 1)
        self.html_tag = "span"
        self.inner_html = "&nbsp;" * self.num


# Non html components


class TabGroup(Div):
    """
    Displays a tab based on its value. Has a dict of tabs whose keys is the value. A tab is any JustPy component.

    format of dict: {'value1': {'tab': comp1, 'order': number}, 'value2': {'tab': comp2, 'order': number} ...}
    self.tabs - tab dict
    self.animation_next = 'slideInRight'    set animation for tab coming in
    self.animation_prev = 'slideOutLeft'    set animation for tab going out
    self.animation_speed = 'faster'  can be on of  '' | 'slow' | 'slower' | 'fast'  | 'faster'
    self.value  value of group and tab to display
    self.previous - previous tab, no need to change except to set to '' in order to display tab without animation which is default at first

    """

    wrapper_classes = " "
    wrapper_style = "display: flex; position: absolute; width: 100%; height: 100%;  align-items: center; justify-content: center; background-color: #fff;"

    def __init__(self, **kwargs):

        self.tabs = (
            {}
        )  # Dict with format 'value': {'tab': Div component, 'order': number} for each entry
        self.value = ""
        self.previous_value = ""
        # https://github.com/daneden/animate.css
        self.animation_next = "slideInRight"
        self.animation_prev = "slideOutLeft"
        self.animation_speed = "faster"  # '' | 'slow' | 'slower' | 'fast'  | 'faster'

        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        if key == "value":
            try:
                self.previous_value = self.value
            except:
                pass
        self.__dict__[key] = value

    def model_update(self):
        self.value = self.model[0].data[self.model[1]]

    def convert_object_to_dict(self):
        self.components = []
        self.wrapper_div_classes = (
            self.animation_speed
        )  # Component in this will be centered

        if self.previous_value:
            self.wrapper_div = Div(
                classes=self.wrapper_div_classes,
                animation=self.animation_next,
                temp=True,
                style=f"{self.__class__.wrapper_style} z-index: 50;",
                a=self,
            )
            self.wrapper_div.add(self.tabs[self.value]["tab"])
            self.wrapper_div = Div(
                classes=self.wrapper_div_classes,
                animation=self.animation_prev,
                temp=True,
                style=f"{self.__class__.wrapper_style} z-index: 0;",
                a=self,
            )
            self.wrapper_div.add(self.tabs[self.previous_value]["tab"])
        else:
            self.wrapper_div = Div(
                classes=self.wrapper_div_classes,
                temp=True,
                a=self,
                style=self.__class__.wrapper_style,
            )
            self.wrapper_div.add(self.tabs[self.value]["tab"])

        self.style = (
            " position: relative; overflow: hidden; " + self.style
        )  # overflow: hidden;
        d = super().convert_object_to_dict()
        return d


# HTML tags for which corresponding classes will be created
_tag_create_list = [
    "address",
    "article",
    "aside",
    "footer",
    "header",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "main",
    "nav",
    "section",
    "blockquote",
    "dd",
    "dl",
    "dt",
    "figcaption",
    "figure",
    "hr",
    "li",
    "ol",
    "p",
    "pre",
    "ul",
    "abbr",
    "b",
    "bdi",
    "bdo",
    "br",
    "cite",
    "code",
    "data",
    "dfn",
    "em",
    "i",
    "kbd",
    "mark",
    "q",
    "rb",
    "rp",
    "rt",
    "rtc",
    "ruby",
    "s",
    "samp",
    "small",
    "span",
    "strong",
    "sub",
    "sup",
    "time",
    "tt",
    "u",
    "var",
    "wbr",
    "area",
    "audio",
    "img",
    "map",
    "track",
    "video",
    "embed",
    "iframe",
    "object",
    "param",
    "picture",
    "source",
    "del",
    "ins",
    "title",
    "caption",
    "col",
    "colgroup",
    "table",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "tr",
    "button",
    "fieldset",
    "legend",
    "meter",
    "optgroup",
    "option",
    "progress",  # datalist not supported
    "details",
    "summary",
    "style",  # dialog not supported
]

# Only tags that have non-gloabal  attributes that are supported by HTML 5 are in this dict
_attr_dict = {
    "a": ["download", "href", "hreflang", "media", "ping", "rel", "target", "type"],
    "area": [
        "alt",
        "coords",
        "download",
        "href",
        "hreflang",
        "media",
        "rel",
        "shape",
        "target",
        "type",
    ],
    "audio": ["autoplay", "controls", "loop", "muted", "preload", "src"],
    "base": ["href", "target"],
    "bdo": ["dir"],
    "blockquote": ["cite"],
    "button": [
        "autofocus",
        "disabled",
        "form",
        "formaction",
        "formenctype",
        "formmethod",
        "formnovalidate",
        "formtarget",
        "name",
        "type",
        "value",
    ],
    "canvas": ["height", "width"],
    "col": ["span"],
    "colgroup": ["span"],
    "data": ["value"],
    "del": ["cite", "datetime"],
    "details": ["open"],
    "dialog": ["open"],
    "embed": ["height", "src", "type", "width"],
    "fieldset": ["disabled", "form", "name"],
    "form": [
        "accept-charset",
        "action",
        "autocomplete",
        "enctype",
        "method",
        "name",
        "novalidate",
        "target",
    ],
    "html": ["xmlns"],
    "iframe": ["height", "name", "sandbox", "src", "srcdoc", "width"],
    "img": [
        "alt",
        "crossorigin",
        "height",
        "ismap",
        "longdesc",
        "sizes",
        "src",
        "srcset",
        "usemap",
        "width",
    ],
    "input": [
        "accept",
        "alt",
        "autocomplete",
        "autofocus",
        "checked",
        "dirname",
        "disabled",
        "form",
        "formaction",
        "formenctype",
        "formmethod",
        "formnovalidate",
        "formtarget",
        "height",
        "list",
        "max",
        "maxlength",
        "min",
        "minlength",
        "multiple",
        "name",
        "pattern",
        "placeholder",
        "readonly",
        "required",
        "size",
        "src",
        "step",
        "type",
        "value",
        "width",
    ],
    "ins": ["cite", "datetime"],
    "label": ["for", "form"],
    "li": ["value"],
    "link": ["crossorigin", "href", "hreflang", "media", "rel", "sizes", "type"],
    "map": ["name"],
    "meta": ["charset", "content", "http-equiv", "name"],
    "meter": ["form", "high", "low", "max", "min", "optimum", "value"],
    "object": ["data", "form", "height", "name", "type", "usemap", "width"],
    "ol": ["reversed", "start", "type"],
    "optgroup": ["disabled", "label"],
    "option": ["disabled", "label", "selected", "value"],
    "output": ["for", "form", "name"],
    "param": ["name", "value"],
    "progress": ["max", "value"],
    "q": ["cite"],
    "script": ["async", "charset", "defer", "src", "type"],
    "select": ["autofocus", "disabled", "form", "multiple", "name", "required", "size"],
    "source": ["src", "srcset", "media", "sizes", "type"],
    "style": ["media", "type"],
    "td": ["colspan", "headers", "rowspan"],
    "textarea": [
        "autofocus",
        "cols",
        "dirname",
        "disabled",
        "form",
        "maxlength",
        "name",
        "placeholder",
        "readonly",
        "required",
        "rows",
        "wrap",
    ],
    "th": ["abbr", "colspan", "headers", "rowspan", "scope", "sorted"],
    "time": ["datetime"],
    "track": ["default", "kind", "label", "src", "srclang"],
    "video": [
        "autoplay",
        "controls",
        "height",
        "loop",
        "muted",
        "poster",
        "preload",
        "src",
        "width",
    ],
}

# Name definition for static syntax analysers
# Classes are defined dynamically right after, this is just to assist code editors

Address = (
    Article
) = (
    Aside
) = (
    Footer
) = (
    Header
) = (
    H1
) = (
    H2
) = (
    H3
) = (
    H4
) = (
    H5
) = (
    H6
) = (
    Main
) = (
    Nav
) = (
    Section
) = (
    Blockquote
) = (
    Dd
) = (
    Dl
) = (
    Dt
) = (
    Figcaption
) = (
    Figure
) = (
    Hr
) = (
    Li
) = (
    Ol
) = (
    P
) = (
    Pre
) = (
    Ul
) = (
    Abbr
) = (
    B
) = (
    Bdi
) = (
    Bdo
) = (
    Br
) = (
    Cite
) = (
    Code
) = (
    Data
) = (
    Dfn
) = (
    Em
) = (
    I
) = (
    Kbd
) = (
    Mark
) = (
    Q
) = (
    Rb
) = (
    Rp
) = (
    Rt
) = (
    Rtc
) = (
    Ruby
) = (
    S
) = (
    Samp
) = (
    Small
) = (
    Span
) = (
    Strong
) = (
    Sub
) = (
    Sup
) = (
    Time
) = (
    Tt
) = (
    U
) = (
    Var
) = (
    Wbr
) = (
    Area
) = (
    Audio
) = (
    Img
) = (
    Map
) = (
    Track
) = (
    Video
) = (
    Embed
) = (
    Iframe
) = (
    Object
) = (
    Param
) = (
    Picture
) = (
    Source
) = (
    Del
) = (
    Ins
) = (
    Caption
) = (
    Col
) = (
    Colgroup
) = (
    Table
) = (
    Tbody
) = (
    Td
) = (
    Tfoot
) = (
    Th
) = (
    Thead
) = (
    Tr
) = (
    Button
) = Fieldset = Legend = Meter = Optgroup = Option = Progress = Details = Summary = None
Animate = (
    AnimateMotion
) = (
    AnimateTransform
) = (
    Circle
) = (
    ClipPath
) = (
    Defs
) = (
    Desc
) = (
    Discard
) = (
    Ellipse
) = (
    FeBlend
) = (
    FeColorMatrix
) = (
    FeComponentTransfer
) = (
    FeComposite
) = (
    FeConvolveMatrix
) = (
    FeDiffuseLighting
) = (
    FeDisplacementMap
) = (
    FeDistantLight
) = (
    FeDropShadow
) = (
    FeFlood
) = (
    FeFuncA
) = (
    FeFuncB
) = (
    FeFuncG
) = (
    FeFuncR
) = (
    FeGaussianBlur
) = (
    FeImage
) = (
    FeMerge
) = (
    FeMergeNode
) = (
    FeMorphology
) = (
    FeOffset
) = (
    FePointLight
) = (
    FeSpecularLighting
) = (
    FeSpotLight
) = (
    FeTile
) = (
    FeTurbulence
) = (
    Filter
) = (
    ForeignObject
) = (
    G
) = (
    Image
) = (
    Line
) = (
    LinearGradient
) = (
    Marker
) = (
    Mask
) = (
    Metadata
) = (
    Mpath
) = (
    Path
) = (
    Pattern
) = (
    Polygon
) = (
    Polyline
) = (
    RadialGradient
) = (
    Rect
) = Set = Stop = Svg = Switch = Symbol = Text = TextPath = Tspan = Use = View = None

# Tag classes defined dynamically at runtime
for tag in _tag_create_list:
    globals()[tag.capitalize()] = type(
        tag.capitalize(),
        (Div,),
        {"html_tag": tag, "attributes": _attr_dict.get(tag, [])},
    )

# **********************************
# SVG components
# https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute

# in, in2, mode
svg_tags = [
    "a",
    "animate",
    "animateMotion",
    "animateTransform",
    "audio",
    "canvas",
    "circle",
    "clipPath",
    "defs",
    "desc",
    "discard",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "foreignObject",
    "g",
    "iframe",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "script",
    "set",
    "stop",
    "style",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "title",
    "tspan",
    "unknown",
    "use",
    "video",
    "view",
]

svg_tags_use = [
    "animate",
    "animateMotion",
    "animateTransform",
    "circle",
    "clipPath",
    "defs",
    "desc",
    "discard",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "foreignObject",
    "g",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "set",
    "stop",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "tspan",
    "use",
    "view",
]

svg_presentation_attributes = [
    "alignment-baseline",
    "baseline-shift",
    "clip",
    "clip-path",
    "clip-rule",
    "color",
    "color-interpolation",
    "color-interpolation-filters",
    "color-profile",
    "color-rendering",
    "cursor",
    "direction",
    "display",
    "dominant-baseline",
    "enable-background",
    "fill",
    "fill-opacity",
    "fill-rule",
    "filter",
    "flood-color",
    "flood-opacity",
    "font-family",
    "font-size",
    "font-size-adjust",
    "font-stretch",
    "font-style",
    "font-variant",
    "font-weight",
    "glyph-orientation-horizontal",
    "glyph-orientation-vertical",
    "image-rendering",
    "kerning",
    "letter-spacing",
    "lighting-color",
    "marker-end",
    "marker-mid",
    "marker-start",
    "mask",
    "opacity",
    "overflow",
    "pointer-events",
    "shape-rendering",
    "stop-color",
    "stop-opacity",
    "stroke",
    "stroke-dasharray",
    "stroke-dashoffset",
    "stroke-linecap",
    "stroke-linejoin",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-width",
    "text-anchor",
    "transform",
    "text-decoration",
    "text-rendering",
    "unicode-bidi",
    "vector-effect",
    "visibility",
    "word-spacing",
    "writing-mode",
    "cx",
    "cy",
    "r",
    "rx",
    "ry",
    "d",
    "fill",
    "transform",
]

svg_filter_attributes = [
    "height",
    "result",
    "width",
    "x",
    "y",
    "type",
    "tableValues",
    "slope",
    "intercept",
    "amplitude",
    "exponent",
    "offset",
    "xlink:href",
]

svg_animation_attributes = [
    "attributeType",
    "attributeName",
    "begin",
    "dur",
    "end",
    "min",
    "max",
    "restart",
    "repeatCount",
    "repeatDur",
    "fill",
    "calcMode",
    "values",
    "keyTimes",
    "keySplines",
    "from",
    "to",
    "by",
    "additive",
    "accumulate",
]

svg_attr_dict = {
    "a": ["download", "requiredExtensions", "role", "systemLanguage"],
    "animate": [
        "accumulate",
        "additive",
        "attributeName",
        "begin",
        "by",
        "calcMode",
        "dur",
        "end",
        "fill",
        "from",
        "href",
        "keySplines",
        "keyTimes",
        "max",
        "min",
        "repeatCount",
        "repeatDur",
        "requiredExtensions",
        "restart",
        "systemLanguage",
        "to",
        "values",
    ],
    "animateMotion": [
        "accumulate",
        "additive",
        "begin",
        "by",
        "calcMode",
        "dur",
        "end",
        "fill",
        "from",
        "href",
        "keyPoints",
        "keySplines",
        "keyTimes",
        "max",
        "min",
        "origin",
        "path",
        "repeatCount",
        "repeatDur",
        "requiredExtensions",
        "restart",
        "rotate",
        "systemLanguage",
        "to",
        "values",
    ],
    "animateTransform": [
        "accumulate",
        "additive",
        "attributeName",
        "begin",
        "by",
        "calcMode",
        "dur",
        "end",
        "fill",
        "from",
        "href",
        "keySplines",
        "keyTimes",
        "max",
        "min",
        "repeatCount",
        "repeatDur",
        "requiredExtensions",
        "restart",
        "systemLanguage",
        "to",
        "type",
        "values",
    ],
    "audio": ["requiredExtensions", "role", "systemLanguage"],
    "canvas": ["preserveAspectRatio", "requiredExtensions", "role", "systemLanguage"],
    "circle": ["pathLength", "requiredExtensions", "role", "systemLanguage"],
    "clipPath": ["clipPathUnits", "requiredExtensions", "systemLanguage"],
    "discard": ["begin", "href", "requiredExtensions", "role", "systemLanguage"],
    "ellipse": ["pathLength", "requiredExtensions", "role", "systemLanguage"],
    "feBlend": ["height", "in", "in2", "mode", "result", "width", "x", "y"],
    "feColorMatrix": ["height", "in", "result", "type", "values", "width", "x", "y"],
    "feComponentTransfer": ["height", "in", "result", "width", "x", "y"],
    "feComposite": [
        "height",
        "in",
        "in2",
        "k1",
        "k2",
        "k3",
        "k4",
        "operator",
        "result",
        "width",
        "x",
        "y",
    ],
    "feConvolveMatrix": [
        "bias",
        "divisor",
        "edgeMode",
        "height",
        "in",
        "kernelMatrix",
        "kernelUnitLength",
        "order",
        "preserveAlpha",
        "result",
        "targetX",
        "targetY",
        "width",
        "x",
        "y",
    ],
    "feDiffuseLighting": [
        "diffuseConstant",
        "height",
        "in",
        "kernelUnitLength",
        "result",
        "surfaceScale",
        "width",
        "x",
        "y",
    ],
    "feDisplacementMap": [
        "height",
        "in",
        "in2",
        "result",
        "scale",
        "width",
        "x",
        "xChannelSelector",
        "y",
        "yChannelSelector",
    ],
    "feDistantLight": ["azimuth", "elevation"],
    "feDropShadow": [
        "dx",
        "dy",
        "height",
        "in",
        "result",
        "stdDeviation",
        "width",
        "x",
        "y",
    ],
    "feFlood": ["height", "result", "width", "x", "y"],
    "feFuncA": [
        "amplitude",
        "exponent",
        "intercept",
        "offset",
        "slope",
        "tableValues",
        "type",
    ],
    "feFuncB": [
        "amplitude",
        "exponent",
        "intercept",
        "offset",
        "slope",
        "tableValues",
        "type",
    ],
    "feFuncG": [
        "amplitude",
        "exponent",
        "intercept",
        "offset",
        "slope",
        "tableValues",
        "type",
    ],
    "feFuncR": [
        "amplitude",
        "exponent",
        "intercept",
        "offset",
        "slope",
        "tableValues",
        "type",
    ],
    "feGaussianBlur": [
        "edgeMode",
        "height",
        "in",
        "result",
        "stdDeviation",
        "width",
        "x",
        "y",
    ],
    "feImage": [
        "crossorigin",
        "height",
        "href",
        "preserveAspectRatio",
        "result",
        "width",
        "x",
        "y",
    ],
    "feMerge": ["height", "result", "width", "x", "y"],
    "feMergeNode": ["in"],
    "feMorphology": ["height", "in", "operator", "radius", "result", "width", "x", "y"],
    "feOffset": ["dx", "dy", "height", "in", "result", "width", "x", "y"],
    "fePointLight": ["x", "y", "z"],
    "feSpecularLighting": [
        "height",
        "in",
        "kernelUnitLength",
        "result",
        "specularConstant",
        "specularExponent",
        "surfaceScale",
        "width",
        "x",
        "y",
    ],
    "feSpotLight": [
        "limitingConeAngle",
        "pointsAtX",
        "pointsAtY",
        "pointsAtZ",
        "specularExponent",
        "x",
        "y",
        "z",
    ],
    "feTile": ["height", "in", "result", "width", "x", "y"],
    "feTurbulence": [
        "baseFrequency",
        "height",
        "numOctaves",
        "result",
        "seed",
        "stitchTiles",
        "type",
        "width",
        "x",
        "y",
    ],
    "filter": ["filterUnits", "height", "primitiveUnits", "width", "x", "y"],
    "foreignObject": ["requiredExtensions", "role", "systemLanguage"],
    "g": ["requiredExtensions", "role", "systemLanguage"],
    "iframe": ["requiredExtensions", "role", "systemLanguage"],
    "image": [
        "crossorigin",
        "href",
        "preserveAspectRatio",
        "requiredExtensions",
        "role",
        "systemLanguage",
    ],
    "line": [
        "pathLength",
        "requiredExtensions",
        "role",
        "systemLanguage",
        "x1",
        "x2",
        "y1",
        "y2",
    ],
    "linearGradient": [
        "gradientTransform",
        "gradientUnits",
        "href",
        "spreadMethod",
        "x1",
        "x2",
        "y1",
        "y2",
    ],
    "marker": [
        "markerHeight",
        "markerUnits",
        "markerWidth",
        "orient",
        "preserveAspectRatio",
        "refX",
        "refY",
        "viewBox",
    ],
    "mask": [
        "height",
        "maskContentUnits",
        "maskUnits",
        "requiredExtensions",
        "systemLanguage",
        "width",
        "x",
        "y",
    ],
    "mpath": ["href"],
    "path": ["pathLength", "requiredExtensions", "role", "systemLanguage"],
    "pattern": [
        "height",
        "href",
        "patternContentUnits",
        "patternTransform",
        "patternUnits",
        "preserveAspectRatio",
        "viewBox",
        "width",
        "x",
        "y",
    ],
    "polygon": ["pathLength", "points", "requiredExtensions", "role", "systemLanguage"],
    "polyline": [
        "pathLength",
        "points",
        "requiredExtensions",
        "role",
        "systemLanguage",
    ],
    "radialGradient": [
        "cx",
        "cy",
        "fr",
        "fx",
        "fy",
        "gradientTransform",
        "gradientUnits",
        "href",
        "r",
        "spreadMethod",
    ],
    "rect": ["pathLength", "requiredExtensions", "role", "systemLanguage"],
    "script": ["href"],
    "set": [
        "attributeName",
        "begin",
        "dur",
        "end",
        "fill",
        "href",
        "max",
        "min",
        "repeatCount",
        "repeatDur",
        "requiredExtensions",
        "restart",
        "systemLanguage",
        "to",
    ],
    "stop": ["offset"],
    "style": ["media"],
    "svg": [
        "playbackorder",
        "preserveAspectRatio",
        "requiredExtensions",
        "role",
        "systemLanguage",
        "timelinebegin",
        "transform",
        "viewBox",
        "zoomAndPan",
        "xmlns",
        "version",
    ],
    "switch": ["requiredExtensions", "role", "systemLanguage"],
    "symbol": ["preserveAspectRatio", "refX", "refY", "role", "viewBox"],
    "text": [
        "dx",
        "dy",
        "lengthAdjust",
        "requiredExtensions",
        "role",
        "rotate",
        "systemLanguage",
        "textLength",
        "x",
        "y",
    ],
    "textPath": [
        "href",
        "lengthAdjust",
        "method",
        "path",
        "requiredExtensions",
        "role",
        "side",
        "spacing",
        "startOffset",
        "systemLanguage",
        "textLength",
    ],
    "tspan": [
        "dx",
        "dy",
        "lengthAdjust",
        "requiredExtensions",
        "role",
        "rotate",
        "systemLanguage",
        "textLength",
        "x",
        "y",
    ],
    "unknown": ["requiredExtensions", "role", "systemLanguage"],
    "use": ["href", "requiredExtensions", "role", "systemLanguage"],
    "video": ["requiredExtensions", "role", "systemLanguage"],
    "view": ["preserveAspectRatio", "role", "viewBox", "zoomAndPan"],
}

for tag in svg_tags_use:
    c_tag = tag[0].capitalize() + tag[1:]
    globals()[c_tag] = type(
        c_tag,
        (Div,),
        {
            "html_tag": tag,
            "attributes": svg_attr_dict.get(tag, [])
            + svg_presentation_attributes
            + svg_filter_attributes,
        },
    )


# *************************** end SVG components


class HTMLEntity(Span):
    # Render HTML Entities

    def __init__(self, **kwargs):
        self.entity = ""
        super().__init__(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["inner_html"] = self.entity
        return d


class Hello(Div):
    def __init__(self, **kwargs):
        self.counter = 1
        super().__init__(**kwargs)
        self.classes = "m-1 p-1 text-2xl text-center text-white bg-blue-500 hover:bg-blue-800 cursor-pointer"
        self.text = "Hello! (click me)"

        async def click(self, msg):
            self.text = f"Hello! I was clicked {self.counter} times"
            self.counter += 1

        self.on("click", click)


class QHello(Hello):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.classes = "text-h3 text-primary q-ma-md"


def component_by_tag(tag, attrs=[], **kwargs):
    # tag = tag.lower()
    if tag[0:2] == "q-":
        if tag in _tag_class_dict:
            c = _tag_class_dict[tag](**kwargs)
        else:
            raise ValueError(f"Tag not defined: {tag}")
    elif tag in JustPy.component_registry:
        attributes = JustPy.component_registry[tag]["attributes"]
        attr_dict = {}
        for attr in attrs:
            if attr[0] in attributes:
                attr_dict[attr[0]] = attr[1]
        c = JustPy.component_registry[tag]["class"](**attr_dict)
    else:
        tag_class_name = tag[0].capitalize() + tag[1:]
        try:
            c = globals()[tag_class_name](**kwargs)
        except:
            raise ValueError(f"Tag not defined: {tag}")
    return c


class AutoTable(Table):
    """
    Creates an HTML table from a list of lists
    First list is used as headers
    """

    td_classes = "border px-4 py-2 text-center"
    tr_even_classes = "bg-gray-100 "
    tr_odd_classes = ""
    th_classes = "px-4 py-2"

    def __init__(self, **kwargs):
        self.values = []
        super().__init__(**kwargs)

    def react(self, data):
        self.set_class("table-auto")
        # First row of values is header
        if self.values:
            headers = self.values[0]
            thead = Thead(a=self)
            tr = Tr(a=thead)
            for item in headers:
                Th(text=item, classes=self.th_classes, a=tr)
            tbody = Tbody(a=self)
            for i, row in enumerate(self.values[1:]):
                if i % 2 == 1:
                    tr = Tr(classes=self.tr_even_classes, a=tbody)
                else:
                    tr = Tr(classes=self.tr_odd_classes, a=tbody)
                for item in row:
                    Td(text=item, classes=self.td_classes, a=tr)


get_tag = component_by_tag


class BasicHTMLParser(HTMLParser):
    # Void elements do not need closing tag
    void_elements = [
        "area",
        "base",
        "br",
        "col",
        "embed",
        "hr",
        "img",
        "input",
        "keygen",
        "link",
        "menuitem",
        "meta",
        "param",
        "source",
        "track",
        "wbr",
    ]

    def __init__(self, context, **kwargs):
        super().__init__()
        self.context = context
        self.level = -1
        self.parse_id = 0
        self.start_tag = True
        self.components = []
        self.name_dict = Dict()  # After parsing holds a dict with named components
        self.dict_attribute = kwargs.get(
            "dict_attribute", "name"
        )  # Use another attribute than name
        self.root = Div(name="root")
        self.containers = []
        self.containers.append(self.root)
        self.endtag_required = True
        self.create_commands = kwargs.get(
            "create_commands", True
        )  # If True, create the justpy command list
        self.command_prefix = kwargs.get(
            "command_prefix", "jp."
        )  # Prefix for commands generated, defaults to 'jp.'
        if self.create_commands:
            # List of command strings (justpy python code to generate the element)
            self.commands = [f"root = {self.command_prefix}Div()"]
        else:
            self.commands = ""

    def parse_starttag(self, i):
        # This is the original library method with two changes to stop tags and attributes being lower case
        # This is required for the SVG tags which can be camelcase
        # https://github.com/python/cpython/blob/3.7/Lib/html/parser.py
        self.__starttag_text = None
        endpos = self.check_for_whole_start_tag(i)
        if endpos < 0:
            return endpos
        rawdata = self.rawdata
        self.__starttag_text = rawdata[i:endpos]

        # Now parse the data between i+1 and j into a tag and attrs
        attrs = []
        match = tagfind_tolerant.match(rawdata, i + 1)
        assert match, "unexpected call to parse_starttag()"
        k = match.end()
        # self.lasttag = tag = match.group(1).lower() was the original
        self.lasttag = tag = match.group(1)
        while k < endpos:
            m = attrfind_tolerant.match(rawdata, k)
            if not m:
                break
            attrname, rest, attrvalue = m.group(1, 2, 3)
            if not rest:
                attrvalue = None
            elif (
                attrvalue[:1] == "'" == attrvalue[-1:]
                or attrvalue[:1] == '"' == attrvalue[-1:]
            ):
                attrvalue = attrvalue[1:-1]
            if attrvalue:
                attrvalue = unescape(attrvalue)
            # attrs.append((attrname.lower(), attrvalue)) was the original
            attrs.append((attrname, attrvalue))
            k = m.end()

        end = rawdata[k:endpos].strip()
        if end not in (">", "/>"):
            lineno, offset = self.getpos()
            if "\n" in self.__starttag_text:
                lineno = lineno + self.__starttag_text.count("\n")
                offset = len(self.__starttag_text) - self.__starttag_text.rfind("\n")
            else:
                offset = offset + len(self.__starttag_text)
            self.handle_data(rawdata[i:endpos])
            return endpos
        if end.endswith("/>"):
            # XHTML-style empty tag: <span attr="value" />
            self.handle_startendtag(tag, attrs)
        else:
            self.handle_starttag(tag, attrs)
            if tag in self.CDATA_CONTENT_ELEMENTS:
                self.set_cdata_mode(tag)
        return endpos

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs)
        if self.endtag_required:
            self.handle_endtag(tag)
        else:
            self.endtag_required = True

    def handle_starttag(self, tag, attrs):
        self.level = self.level + 1
        self.parse_id += 1
        c = component_by_tag(tag, attrs)
        c.parse_id = self.parse_id
        command_string = f""
        if c is None:
            print(
                tag,
                "No such tag, Div being used instead *****************************************",
            )
            c = Div()
        for attr in attrs:
            attr = list(attr)
            attr[0] = attr[0].replace("-", "_")
            if attr[0][0] == "@":
                if attr[1] in self.context.f_locals:
                    c.on(attr[0][1:], self.context.f_locals[attr[1]])
                elif attr[1] in self.context.f_globals:
                    c.on(attr[0][1:], self.context.f_globals[attr[1]])
                else:
                    cls = JustpyBaseComponent
                    if not c.id:
                        c.id = cls.next_id
                        cls.next_id += 1
                    fn_string = f"def oneliner{c.id}(self, msg):\n {attr[1]}"  # remove first and last characters which are quotes
                    exec(fn_string)
                    c.on(attr[0][1:], locals()[f"oneliner{c.id}"])
                continue
            if attr[0][0] == ":":
                attr[0] = attr[0][1:]
                attr[1] = eval(attr[1])
            if attr[0] == "id":
                c.id = attr[1]
                continue
            if attr[1] is None:
                setattr(c, attr[0], True)
                attr[1] = True
            else:
                setattr(c, attr[0], attr[1])
            # Add to name to dict of named components. Each entry can be a list of components to allow multiple components with same name
            if attr[0] == self.dict_attribute:
                if attr[1] not in self.name_dict:
                    self.name_dict[attr[1]] = c
                else:
                    if not isinstance(self.name_dict[attr[1]], (list,)):
                        self.name_dict[attr[1]] = [self.name_dict[attr[1]]]
                    self.name_dict[attr[1]].append(c)
            if attr[0] == "class":
                c.classes = attr[1]
                attr[0] = "classes"
            # Handle attributes that are also python reserved words
            if attr[0] in ["in", "from"]:
                attr[0] = "_" + attr[0]

            if self.create_commands:
                if isinstance(attr[1], str):
                    command_string = f"{command_string}{attr[0]}='{attr[1]}', "
                else:
                    command_string = f"{command_string}{attr[0]}={attr[1]}, "

        if self.create_commands:
            if id(self.containers[-1]) == id(self.root):
                command_string = f"c{c.parse_id} = {self.command_prefix}{c.class_name}({command_string}a=root)"
            else:
                command_string = f"c{c.parse_id} = {self.command_prefix}{c.class_name}({command_string}a=c{self.containers[-1].parse_id})"
            self.commands.append(command_string)

        self.containers[-1].add_component(c)
        self.containers.append(c)

        if tag in BasicHTMLParser.void_elements:
            self.handle_endtag(tag)
            self.endtag_required = False
        else:
            self.endtag_required = True

    def handle_endtag(self, tag):
        c = self.containers.pop()
        del c.parse_id
        self.level = self.level - 1

    def handle_data(self, data):
        data = data.strip()
        if data:
            self.containers[-1].text = data
            data = data.replace("'", "\\'")
            if self.create_commands:
                self.commands[-1] = f"{self.commands[-1][:-1]}, text='{data}')"
        return

    def handle_comment(self, data):
        pass

    def handle_entityref(self, name):
        c = chr(name2codepoint[name])

    def handle_charref(self, name):
        if name.startswith("x"):
            c = chr(int(name[1:], 16))
        else:
            c = chr(int(name))

    def handle_decl(self, data):
        pass


def justPY_parser(html_string, context, **kwargs):
    """
    Returns root component of the parser with the name_dict as attribute.
    If root component has only one child, returns the child
    """
    parser = BasicHTMLParser(context, **kwargs)
    parser.feed(html_string)
    if len(parser.root.components) == 1:
        parser_result = parser.root.components[0]
    else:
        parser_result = parser.root
    parser_result.name_dict = parser.name_dict
    parser_result.commands = parser.commands
    parser_result.initialize(**kwargs)
    return parser_result


def parse_html(html_string, **kwargs):
    return justPY_parser(html_string, inspect.stack()[1][0], **kwargs)


def parse_html_file(html_file, **kwargs):
    with open(html_file, encoding="utf-8") as f:
        return justPY_parser(f.read(), inspect.stack()[1][0], **kwargs)


try:
    import aiofiles

    _has_aiofiles = True
except:
    _has_aiofiles = False

if _has_aiofiles:

    async def parse_html_file_async(html_file, **kwargs):
        async with aiofiles.open(html_file, encoding="utf-8") as f:
            s = await f.read()
        return justPY_parser(s, **kwargs)

else:

    async def parse_html_file_async(html_file, **kwargs):
        raise Exception("aiofiles not installed")


async def get(url, format="json"):
    async with httpx.AsyncClient() as client:
        result = await client.get(url)
    if format == "json":
        return result.json()
    else:
        return result.text


try:
    import markdown

    _has_markdown = True
except:
    _has_markdown = False


if _has_markdown:

    class Markdown(Div):

        all_markdown_extensions = [
            "extra",
            "abbr",
            "attr_list",
            "def_list",
            "fenced_code",
            "footnotes",
            "md_in_html",
            "tables",
            "admonition",
            "codehilite",
            "legacy_attrs",
            "legacy_em",
            "meta",
            "nl2br",
            "sane_lists",
            "smarty",
            "toc",
            "wikilinks",
        ]

        def __init__(self, **kwargs):
            self.markdown = ""
            self._cache = ""
            self.all_extensions = True
            self._html_result = ""
            self.inner_html = ""
            self.extensions = []
            super().__init__(**kwargs)

        def convert_object_to_dict(self):
            d = super().convert_object_to_dict()
            if "prose" not in d["classes"].split():
                d["classes"] += " prose"
            # Use all extensions if all_extensions is True and no extensions were specified
            if self.all_extensions and not self.extensions:
                self.extensions = self.all_markdown_extensions
            if self.markdown != self._cache:
                d["inner_html"] = self._html_result = markdown.markdown(
                    self.markdown, extensions=self.extensions
                )
                self._cache = self.markdown
            else:
                d["inner_html"] = self._html_result
            return d


class Equation(Div):

    vue_type = "katexjp"

    def __init__(self, **kwargs):
        self.equation = ""
        self.options = {"macros": {}}
        kwargs["temp"] = False  # Force an id to be assigned
        super().__init__(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["equation"] = self.equation
        self.options["throwOnError"] = False
        d["options"] = self.options
        return d


def get_websocket(event_data):
    return WebPage.sockets[event_data["page_id"]][event_data["websocket_id"]]


def create_transition():
    return Dict(
        {
            "enter": "",
            "enter_start": "",
            "enter_end": "",
            "leave": "",
            "leave_start": "",
            "leave_end": "",
            "load": "",
            "load_start": "",
            "load_end": "",
        }
    )


class Styles:
    button_simple = (
        "bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded"
    )
    button_pill = (
        "bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full"
    )
    button_outline = "bg-transparent hover:bg-blue-500 text-blue-700 font-semibold hover:text-white py-2 px-4 border border-blue-500 hover:border-transparent rounded"
    button_bordered = "bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 border border-blue-700 rounded"
    button_disabled = "bg-blue-500 text-white font-bold py-2 px-4 rounded opacity-50 cursor-not-allowed"
    button_3d = "bg-blue-500 hover:bg-blue-400 text-white font-bold py-2 px-4 border-b-4 border-blue-700 hover:border-blue-500 rounded"
    button_elevated = "bg-white hover:bg-gray-100 text-gray-800 font-semibold py-2 px-4 border border-gray-400 rounded shadow"

    input_classes = "m-2 bg-gray-200 border-2 border-gray-200 rounded w-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"

    # https://www.lipsum.com /
    lorem_ipsum = """
    Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
    """



# Justpy Tutorial demo women_majors2 
# generated by demoapp.update_from tutorial at 2022-09-21T08:41:18.058714+00:00 
# see https://github.com/justpy-org/justpy/blob/master/examples/charts_tutorial/pandas/women_majors2.py
# 
# Customizing the Chart            
# see https://justpy.io/charts_tutorial/pandas#customizing-the-chart
import justpy as jp
import pandas as pd


wm = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)
wm_under_20 = list(
  wm.loc[0, wm.loc[0] < 20].index
) # Create list of majors which start under 20%

def women_majors2():
    wp = jp.WebPage()

    # First chart
    wm.jp.plot(0, wm_under_20, kind='spline', a=wp, title='The gender gap is transitory - even for extreme cases',
                subtitle='Percentage of Bachelors conferred to women form 1970 to 2011 in the US for extreme cases where the percentage was less than 20% in 1970',
                classes='m-2 p-2 w-3/4')

    # Second Chart
    wm_chart = wm.jp.plot(0, wm_under_20, kind='spline', a=wp, categories=False,
                          title='The gender gap is transitory - even for extreme cases',
                          subtitle='Percentage of Bachelors conferred to women form 1970 to 2011 in the US for extreme cases where the percentage was less than 20% in 1970',
                          classes='m-2 p-2 w-3/4 border', style='height: 700px')
    o = wm_chart.options
    o.title.align = 'left'
    o.title.style.fontSize = '24px'
    o.subtitle.align = 'left'
    o.subtitle.style.fontSize = '20px'
    o.xAxis.title.text = 'Year'
    o.xAxis.gridLineWidth = 1
    o.yAxis.title.text = '% Women in Major'
    o.yAxis.labels.format = '{value}%'
    o.legend.layout = 'proximate'
    o.legend.align = 'right'
    o.plotOptions.series.marker.enabled = False
    return wp

# initialize the demo
from examples.basedemo import Demo
Demo("women_majors2", women_majors2)



"""
Created on 2022-08-27

@author: wf
"""
import justpy as jp


class ClickDemo:
    """
    demo for click handling
    """

    async def onDivClick(self, msg):
        """
        handle a click on the Div
        """
        print(msg)
        self.clickCount += 1
        msg.target.text = f"I was clicked {self.clickCount} times"

    async def click_demo(self):
        """
        the example Webpage under test
        """
        wp = jp.WebPage(debug=True)
        self.clickCount = 0
        d = jp.Div(
            text="Not clicked yet",
            a=wp,
            classes="w-48 text-xl m-2 p-1 bg-blue-500 text-white",
        )
        d.on("click", self.onDivClick)
        d.additional_properties = [
            "screenX",
            "pageY",
            "altKey",
            "which",
            "movementX",
            "button",
            "buttons",
        ]
        return wp


async def click_demo():
    clickDemo = ClickDemo()
    return await clickDemo.click_demo()


from examples.basedemo import Demo

Demo("click demo", click_demo)



# Justpy Tutorial demo input_demo_model1 from docs/tutorial/model_and_data.md
import justpy as jp

async def input_demo_model1(request):
    wp = jp.WebPage(data={ 'text': 'Initial text'})
    input_classes = "m-2 bg-gray-200 appearance-none border-2 border-gray-200 rounded xtw-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"
    jp.Input(a=wp, classes=input_classes, placeholder='Please type here', model=[wp, 'text'])
    jp.Div(model=[wp, 'text'], classes='m-2 p-2 h-32 text-xl border-2 overflow-auto', a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_demo_model1",input_demo_model1)



# Justpy Tutorial demo html_comps3 from docs/tutorial/html_components.md
import justpy as jp

def html_comps3():
    wp = jp.WebPage()
    jp.Div(text='Text in italic', a=wp, classes='italic')
    jp.Div(text='Text in bold', a=wp, classes='font-bold')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("html_comps3",html_comps3)



# Justpy Tutorial demo table_test from docs/tutorial/custom_components.md
import justpy as jp
import pandas as pd


class AutoTable(jp.Table):

    td_classes = 'border px-4 py-2 text-center'
    tr_even_classes = 'bg-gray-100 '
    tr_odd_classes = ''
    th_classes = 'px-4 py-2'

    def __init__(self, **kwargs):
        self.values = []
        super().__init__(**kwargs)


    def react(self,data):
        self.set_class('table-auto')
        #First row of values is header
        if self.values:
            headers = self.values[0]
            thead = jp.Thead(a=self)
            tr = jp.Tr(a=thead)
            for item in headers:
                jp.Th(text=item, classes=self.th_classes, a=tr)
            tbody = jp.Tbody(a=self)
            for i, row in enumerate(self.values[1:]):
                if i % 2 == 1:
                    tr = jp.Tr(classes=self.tr_even_classes, a=tbody)
                else:
                    tr = jp.Tr(classes=self.tr_odd_classes, a=tbody)
                for item in row:
                    jp.Td(text=item, classes=self.td_classes, a=tr)


wm_df = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)
wm_df['Year'] = wm_df['Year'].astype('str')
headers = list(wm_df.columns)
table_data = wm_df.to_numpy().tolist()
table_data.insert(0, headers)

def table_test():
    wp = jp.WebPage()
    d = jp.Div(classes='w-7/8 m-2 p-3 border rounded-lg ', a=wp)
    AutoTable(values=table_data, a=d, classes='block p-4 overflow-auto', style='height: 90vh')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("table_test",table_test)



import justpy as jp
import sqlite3
import pandas as pd
import os
from jpcore.download import Download

# Download the database file to the $HOME/.justpy directory
# from: https://elimintz.github.io/chinook.db, originally from https://www.sqlitetutorial.net/

dbname = "chinook.db"
url = f"https://elimintz.github.io/{dbname}"
os.makedirs(Download.get_cache_path(), exist_ok=True)
filePath = f"{Download.get_cache_path()}/{dbname}"
Download.download_file(url, dbname, target_directory=Download.get_cache_path())

db_con = sqlite3.connect(filePath)
table_names = [
    "albums",
    "artists",
    "customers",
    "sqlite_sequence",
    "employees",
    "genres",
    "invoices",
    "invoice_items",
    "media_types",
    "playlists",
    "playlist_track",
    "tracks",
    "sqlite_stat1",
]

tables = {}
for table_name in table_names:
    tables[table_name] = pd.read_sql_query(f"SELECT * from {table_name}", db_con)


def selected_event(self, msg):
    # Runs when a table name is selected
    # Create a new grid and use its column and row definitions for grid already on page
    new_grid = tables[msg.value].jp.ag_grid(temp=True)
    msg.page.g.options.columnDefs = new_grid.options.columnDefs
    msg.page.g.options.rowData = new_grid.options.rowData


def db_test(request):
    wp = jp.QuasarPage()
    table_name = request.query_params.get("table", "albums")
    s = jp.QSelect(
        options=table_names,
        a=wp,
        label="Select Table",
        outlined=True,
        input=selected_event,
        style="width: 350px; margin: 0.25rem; padding: 0.25rem;",
        value=table_name,
    )
    g = tables[table_name].jp.ag_grid(
        a=wp, style="height: 90vh; width: 99%; margin: 0.25rem; padding: 0.25rem;"
    )
    g.options.pagination = True
    g.options.paginationAutoPageSize = True
    wp.g = g
    return wp


@jp.SetRoute("/city")
def city_test():
    wp = jp.WebPage()
    g = pd.read_sql_query(
        "SELECT DISTINCT city, country, customerid from customers ORDER BY country",
        db_con,
    ).jp.ag_grid(a=wp)
    g.style = "height: 99vh; width: 450px; margin: 0.25rem; padding: 0.25rem;"
    return wp


from examples.basedemo import Demo

Demo("database test", db_test)
